#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <pthread.h>

#include "shmac/shmac.h"
#include "shmac/hw_timer.h"
#include "shmac/timer.h"
#include "shmac/prof.h"
#include "shmac/irq.h"

#include "timer.h"
#include "sched/sched.h"
#include "global_tq.h"
#include "core.h"
#include "core_tq.h"
#include "core_mq.h"
#include "worker.h"
#include "event_log.h"
#include "join_tq.h"
#include "task.h"

#include "context.h"

#define ID (*TILEREG_CPUID)

extern int __malloc_locked();

static int is_idle_task(task_t *my_task)
{
	return my_task->attrs.schedparam.sched_priority == 0;
}

/*
 * Task cancelation
 */

static int should_cancel(task_t *my_task)
{
	return my_task->cancel_state == PTHREAD_CANCEL_ENABLE &&
		   my_task->state == CANCEL_PENDING;
}

static void cancel(irq_regs_t *registers, task_t *my_task)
{
	// pass task as parameter
	registers->r[REG_R0] = (int)my_task;

	// jump to finish_task
	registers->r[REG_PC] = (int)(&finish_task + (ptrdiff_t)4);

	log_task(TASK_CANCELED, my_task);
}

/*
 * Register save/restore
 */

static void save_registers(task_t *task, irq_regs_t *registers)
{
	for (int i = 0; i < ARM_REG_COUNT; i++) {
		task->register_set.r[i] = registers->r[i];
	}
}
static void restore_registers(task_t *task, irq_regs_t *registers)
{
	for (int i = 0; i < ARM_REG_COUNT; i++) {
		registers->r[i] = task->register_set.r[i];
	}
}


/*
 * Context switching
 */

static int retire_current_task(irq_regs_t *registers, task_t *my_task)
{
	if (my_task->blockable && my_task->cond_waiting == 1)
	{
		save_registers(my_task, registers);

		core_tq_block_current(NONE,
			my_task->cond_wait_timeout > 0 ?
			BLOCK_COND_TIME : BLOCK_COND);

		sched_task_retire_hook(my_task);
		return 1;
	}
	else if (my_task->blockable && my_task->barrier_waiting)
	{
		save_registers(my_task, registers);

		core_tq_block_current(NONE, BLOCK_BARRIER);

		sched_task_retire_hook(my_task);
		return 1;
	}
	else if (my_task->join_waiting == 1)
	{
		save_registers(my_task, registers);
		core_tq_remove_current();
		join_block_task(my_task);
		log_taskf(TASK_JOIN_BLOCKED, my_task, 1, *my_task->waits_for->thread);

		sched_task_retire_hook(my_task);

		return 1;
	}
	else if (join_blocked_any())
	{
		save_registers(my_task, registers);
		sched_task_retire_hook(my_task);

		return 1;
	}
	else if (my_task->state == FINISHED ||
		my_task->state == REAPED ||
		should_context_switch())
	{
		save_registers(my_task, registers);
		sched_task_retire_hook(my_task);
		return 1;
	}
	else
	{
		sched_task_continue_hook(my_task);
		return 0;
	}
}

static void setup_initial_call_frame(task_t *task, irq_regs_t *registers)
{
	// clear all registers for good measure
	for (int i = 0; i < ARM_REG_COUNT; i++) {
		registers->r[i] = 0;
	}
	// status register: user mode, no status flags set
	registers->r[REG_SPSR] = 0x10;

	// pass task as first argument to execute_task. we
	// expect it to look in r0 for that argument.
	registers->r[REG_R0] = (int)task;

	// start executing this function. +4 is because irq handler
	// subtracts 4 before jumping back to user mode
	registers->r[REG_PC] = (int)(&execute_task + (ptrdiff_t)4);

	// calculate top of stack from stackaddr attribute. needs
	// to be properly aligned
	int ptr = (int)(task->attrs.stackaddr + (ptrdiff_t)(task->attrs.stacksize - 1));
	ptr &= ~7; // Newlib really neeeds this to be 8-byte-aligned

	// set up the stack
	registers->r[REG_FP] = ptr;
	registers->r[REG_SP] = ptr;
}

static void setup_initial_task_state(task_t *task)
{
	task->epoch_count		= 0;
}

static void setup_task_state(task_t *task, irq_regs_t *registers)
{
	// if this is the first time we're scheduling this task,
	// we need to create a call frame and initialize state.
	// otherwise, just restore registers.

	if (task->epoch_count > 0) {

		// this task has been preempted before, so just restore its state
		restore_registers(task, registers);
	} else {

		setup_initial_call_frame(task, registers);
		setup_initial_task_state(task);
	}
}

static void flush_finished_task(task_t *my_task)
{
	if (my_task->attrs.detachstate == PTHREAD_CREATE_DETACHED) {

		my_task->state = REAPED;
		log_task(TASK_FINISHED, my_task);

		free_task(my_task);
		core_tq_remove_current();

	} else {
		my_task->state = DEQUEUED;
		log_task(TASK_FINISHED, my_task);

		core_tq_remove_current();
	}
}

static void request_work(task_t *idle)
{
	unsigned int core = idle->epoch_count % (*SYS_CPU_COUNT);
	if (core == ID) {
		core = (core + 1) % (*SYS_CPU_COUNT);
	}

	enum core_state st = get_core_state(core);
	int iter = 0, max_iter = *SYS_CPU_COUNT;
	while (st == NOT_IN_USE || st == NOT_READY || core == ID) {
        core = (core + 1) % (*SYS_CPU_COUNT);
		st = get_core_state(core);

		if (iter++ == max_iter) {
			return;
		}
    }

	core_msg_t msg = {
		.message = WORK_REQUEST,
		.extra = (void*)ID
	};

	logf(MQ_WORK_REQUEST_SENT, 2, (int)msg.extra, core);
	core_mq_push(core, msg);
}

static void pick_from_heap(irq_regs_t *registers)
{
	task_t *my_task = core_tq_update_current();

	// if we're about to idle, ask other cores for work
	if (is_idle_task(my_task)) {
		request_work(my_task);
	}

	setup_task_state(my_task, registers);
	sched_task_start_hook(my_task);
}

static int pick_from_join_blocked(irq_regs_t *registers)
{
	task_t *my_task = join_unblock_task();
	if (my_task == NULL) {
		return 0;
	}

	log_taskf(TASK_JOIN_VIVIFIED, my_task, 1, *my_task->waits_for->thread);

	core_tq_force_current(my_task);
	setup_task_state(my_task, registers);
	sched_task_start_hook(my_task);

	my_task->join_waiting = 0;
	my_task->pinned = 0;

	return 1;
}

static void try_move_work(uint32_t target_core)
{
	task_t *work = NULL;
	if (core_tq_pop_first_inactive(NONE, &work)) {
		core_msg_t msg = {
			.message = WORK_RESPONSE,
			.extra = (void*)work
		};
		core_mq_push(target_core, msg);
		log_taskf(MQ_WORK_RESPONSE_SENT, work, 1, target_core);
	}
}

static void process_mq()
{
	if (core_mq_should_process()) {

		core_msg_t msg;
		while (core_mq_pop(&msg)) {
			switch (msg.message)
			{
				case WORK_REQUEST:
					logf(MQ_WORK_REQUEST_RECEIVED, 1, (int)msg.extra);
					try_move_work((uint32_t)msg.extra);
					break;
				case WORK_RESPONSE:
				{
					task_t *t = (task_t*)msg.extra;
					log_task(MQ_WORK_RESPONSE_RECEIVED, t);

					core_tq_push(ID, t, NONE);
					break;
				}
			}
		}
	}
}

static void join_block_failed()
{
	task_t *me = core_tq_get_current();
	if (me != NULL && me->join_waiting) {
		me->pinned = 0;
		me->join_waiting = 0;
	}
}

void schedule(irq_regs_t *registers)
{
	//if (*TILEREG_CPUID == 0) {
		//prof_enable(1);
	//}

	if (!core_lock(ID, TRYLOCK)) {
		join_block_failed();
		log_task(IRQ_FAILED_CORE_LOCK, core_tq_get_current());
		//if (*TILEREG_CPUID == 0) {
			//prof_enable(0);
		//}
		return;
	}


	log_task(IRQ_STARTED, core_tq_get_current());

	process_mq();
	core_tq_unblock_tasks(NONE);

	task_t *current_task = core_tq_get_current();

	if (current_task == NULL)
	{
		pick_from_heap(registers);
	}
	else if (should_cancel(current_task))
	{
		cancel(registers, current_task);
	}
	else
	{
		sched_task_tick_hook(current_task);

		int switch_task = 0;
		if (current_task->state == FINISHED) {
			flush_finished_task(current_task);
			switch_task = 1;
		} else {
			switch_task = retire_current_task(registers, current_task);
		}

		if (switch_task) {
			if (!pick_from_join_blocked(registers)) {
				pick_from_heap(registers);
			}
		} else {
			// not switching, but request work if continuing idle task
			if (is_idle_task(current_task)) {
				request_work(current_task);
			}
		}
	}

	assert(core_tq_get_current() != NULL);
	assert(*core_tq_get_current()->thread <= global_tq_last_created());
	assert((unsigned int)registers->pc < 500000);

	core_tick(ID);
	core_unlock(ID, 1);

	log_task(IRQ_FINISHED, core_tq_get_current());
	//if (*TILEREG_CPUID == 0) {
		//prof_enable(0);
	//}
}

void schedule_main(irq_regs_t *registers)
{
	pthread_t *tid = malloc(sizeof(pthread_t));
	if (tid == NULL) {
		warn("malloc failed!");
		return;
	}

	sched_entity_t *main_se = task_create(tid, NULL, NULL, NULL,
		0, 1, 1, PTHREAD_CANCEL_ENABLE);

	if (main_se == NULL) {
		warn("malloc failed!");
		return;
	}
	task_t *main_task = main_se->task;

	setup_initial_task_state(main_task);
	main_task->register_set = *registers;

	// we don't need locking as no other threads have started yet
	global_tq_insert(*main_task->thread, main_task);
	core_tq_push(ID, main_task, NONE);
	core_tq_update_current();

	log_taskf(TASK_ASSIGNED, main_task, 2, 0, 1);

	assert(main_task == core_tq_get_current()); // we are returning to expected state
	assert(*main_task->thread == MAX_CPUS); // we got corrrect thread ID
	assert(core_tq_size(ID, SCHED_OTHER) == 1); // idle task

	sched_task_start_hook(main_task);
}
