#include <limits.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include "shmac/shmac.h"
#include "shmac/mutex.h"
#include "shmac/swi.h"
#include "shmac/hw_timer.h"

#include "core.h"
#include "sched.h"
#include "global_tq.h"
#include "data/linked_list.h"
#include "data/ring_buffer.h"

#define ID		(*TILEREG_CPUID)

#define se2t(ts,se) struct task_struct *(ts) = \
    (struct task_struct*)((struct sched_entity*)(se))->task

static core_t runqueue[MAX_CPUS];

// used to order tasks by when they should be reactivated
static int timedwait_comparator(void *task1_se, void *task2_se)
{
	se2t(task1,task1_se);
	se2t(task2,task2_se);
	return task2->cond_wait_timeout - task1->cond_wait_timeout;
}
static int condwait_comparator(void *task1_se, void *task2_se)
{
	se2t(task1,task1_se);
	se2t(task2,task2_se);
	return task2->cond_waiting - task1->cond_waiting;
}
static int barrierwait_comparator(void *task1_se, void *task2_se)
{
	se2t(task1,task1_se);
	se2t(task2,task2_se);
	return task2->barrier_waiting - task1->barrier_waiting;
}

static int (*block_cmps[])(void*,void*) = {
	[BLOCK_COND]		= &condwait_comparator,
	[BLOCK_COND_TIME]	= &timedwait_comparator,
	[BLOCK_BARRIER]		= &barrierwait_comparator,
};

void core_init(unsigned int core_num, int (*comparators[])(void*,void*))
{
	if (core_num >= *SYS_CPU_COUNT) {
		return;
	}

	core_t *core	= &runqueue[core_num];
	core->state		= NOT_READY;

	linked_list_init(&(core->queued_tasks));
	linked_list_init(&(core->joined_tasks));

	core->has_queued_msgs = 0;
	for (int c = 0; c < MAX_CPUS; c++) {
		core->queued_msgs[c].buffer_size = MAX_CPUS;
		ring_buffer_init(&(core->queued_msgs[c]));
	}

	int scheds[] = { SCHED_OTHER, SCHED_RR, SCHED_FIFO };
	for (int i = 0; i < sizeof(scheds)/sizeof(scheds[0]); i++) {
		core->queue[scheds[i]].data		= NULL;
		core->queue[scheds[i]].size		= 0;
		core->queue[scheds[i]].next_idx	= 0;
		core->queue[scheds[i]].cmp		= comparators[scheds[i]];
		prio_array_init(&(core->queue[scheds[i]]));
	}

	core->active_queue			= &(core->queue[SCHED_OTHER]);

	for (int r = BLOCK_REASON_MIN; r <= BLOCK_REASON_MAX; r++) {
		core->blocked_tasks[r].data		= NULL;
		core->blocked_tasks[r].size		= 0;
		core->blocked_tasks[r].next_idx	= 0;
		core->blocked_tasks[r].cmp		= block_cmps[r];
		prio_array_init(&(core->blocked_tasks[r]));
	}

	core->current_task			= NULL;
	core->epoch_count			= 0;

	core->lock_object			= SHMAC_MUTEX_INIT_VALUE;
}

core_t *core_get(unsigned int core_num)
{
	return &runqueue[core_num];
}

int core_inited(unsigned int core_num)
{
	if (core_num < *SYS_CPU_COUNT) {
		return runqueue[core_num].state != NOT_READY;
	}
	return 0;
}

unsigned int core_epoch_count(unsigned int core_num)
{
	return runqueue[core_num].epoch_count;
}

void core_tick(unsigned int core_num)
{
	runqueue[core_num].epoch_count++;
}


enum core_state get_core_state(unsigned int core_num)
{
	return core_num < *SYS_CPU_COUNT
		? runqueue[core_num].state
		: NOT_IN_USE;

}

enum core_state get_my_state()
{
    return get_core_state(ID);
}

void set_core_state(unsigned int core_num, enum core_state state)
{
    if (core_num < *SYS_CPU_COUNT)
    {
        runqueue[core_num].state = state;
    }
}

void set_my_state(enum core_state state)
{
    set_core_state(ID, state);
}

int core_lock(unsigned int core_num, enum lock_mode lm)
{
	switch (lm) {
		case LOCK:
			lock_mutex(&(runqueue[core_num].lock_object));
			return 1;
			break;
		case TRYLOCK:
			return trylock_mutex(&(runqueue[core_num].lock_object));
			break;
		case NONE:
			return 0;
	}
	return 0;
}

void core_unlock(unsigned int core_num, int locked)
{
	if (locked == 1) {
		unlock_mutex(&(runqueue[core_num].lock_object));
	}
}
