#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include "shmac/shmac.h"
#include "shmac/swi.h"

#include "global_tq.h"
#include "data/linked_list.h"
#include "data/ring_buffer.h"

#include "sched/sched.h"
#include "core.h"
#include "core_tq.h"

#define ID		(*TILEREG_CPUID)

static void push_residual_tasks();
static void update_active_queue(core_t *core);

static int task_specific(void *se, void *prop) {
	sched_entity_t *_se = (sched_entity_t*)se;
	return _se->task == prop;
}

static int should_pop_cond_waiting(void *se) {
	sched_entity_t *_se = (sched_entity_t*)se;
	return _se != NULL && _se->task->cond_waiting == 0;
}
static int should_pop_cond_time_waiting(void *se) {
	sched_entity_t *_se = (sched_entity_t*)se;
	return _se != NULL && (
		_se->task->cond_waiting == 0 ||
		_se->task->cond_wait_timeout < *SYS_TICKS);
}
static int should_pop_barrier_waiting(void *se) {
	sched_entity_t *_se = (sched_entity_t*)se;
	return _se != NULL && _se->task->barrier_waiting == 0;
}


void core_tq_push(unsigned int core_num, task_t *item, enum lock_mode lm)
{
	core_t *core = core_get(core_num);
	int lock = core_lock(core_num, lm);

	sched_entity_t *se = item->se;

	sched_task_pre_push_hook(core, item);

	if (prio_array_push(&(core->queue[item->attrs.schedpolicy]), se)) {

		// in case of work moving from core to core, dont update these fields
		if (item->state == QUEUED) {
			item->se->start = *SYS_TICKS;
			item->state = RUNNING;
		}

		item->cpu_no	= core_num;
	} else {
		// ... otherwise queue it for later execution
		linked_list_push(&(core->queued_tasks), (void*)se);
	}

	set_core_state(core_num, BUSY);
	core_unlock(core_num, lock);
}

int core_tq_pop_first_inactive(enum lock_mode lm, task_t **popped_task)
{
	int task_is_unpinned(void *arg, void *_)
	{
		return !(((sched_entity_t*)arg)->task->pinned);
	}

	// used for transfering work from one core to another. if
	// a task is found that is not currently executing, remove it
	// from the heap and return it

	core_t *me = core_get(ID);
	int retval = 0;
	int lock = core_lock(ID, lm);
	int threshold = me->current_task == NULL ? 1 : 0;

	sched_entity_t *found = NULL;

	int prios[] = { SCHED_FIFO, SCHED_RR, SCHED_OTHER };
	for (int i = 0; i < sizeof(prios)/sizeof(prios[0]); i++) {

		prio_array_t *q = &(me->queue[prios[i]]);

		// make sure the idle task is never attempted removed, so only
		// remove unpinned tasks (the idle task is the only one that is pinned)
		if (prio_array_size(q) > threshold) {
			if (prio_array_pop_first(q, &task_is_unpinned, NULL, (void**)&found)) {
				*popped_task = found->task;
				retval = 1;
				break;
			}
		}

	}

	if (retval == 1) {
		push_residual_tasks();
	}

	core_unlock(ID, lock);
	return retval;
}


task_t *core_tq_update_current()
{
	core_t *me = core_get(ID);
	update_active_queue(me);

	// simple case: we have to task, so just pop one from the active queue
	if (me->current_task == NULL) {
		me->current_task = ((sched_entity_t*)prio_array_pop(me->active_queue))->task;
		push_residual_tasks();
		return me->current_task;
	}

	prio_array_t *old_task_queue = &me->queue[me->current_task->attrs.schedpolicy];

	// we haven't switched task queues, so maybe do a combined push/pop?
	if (me->active_queue == old_task_queue)
	{
		if (prio_array_size(me->active_queue) > 0 &&
				!prio_array_beats_root(me->active_queue, me->current_task->se)) {
			me->current_task = ((sched_entity_t*)prio_array_pop_push(
						old_task_queue, me->current_task->se))->task;

			assert(me->current_task != NULL);
		}
	}
	else
	{
		prio_array_push(old_task_queue, me->current_task->se);
		me->current_task = ((sched_entity_t*)prio_array_pop(me->active_queue))->task;
		assert(me->current_task != NULL);

		push_residual_tasks();
	}

	return me->current_task;
}

void core_tq_force_current(task_t *new_current)
{
	core_t *me = core_get(ID);

	if (me->current_task != NULL) {
		task_t *old_current = me->current_task;
		me->current_task = new_current;
		prio_array_push(&me->queue[old_current->attrs.schedpolicy], old_current->se);
	} else {
		me->current_task = new_current;
	}
}


void core_tq_update_running_task(enum lock_mode lm, task_t *task, int old_policy)
{

	int lock = core_lock(task->cpu_no, lm);
	core_t *core = core_get(task->cpu_no);

	prio_array_t *q_src = &(core->queue[old_policy]);
	prio_array_t *q_dst = &(core->queue[task->attrs.schedpolicy]);

	sched_entity_t *found = NULL;
	if (prio_array_pop_first(q_src, &task_specific, (void*)task, (void**)&found)) {
		prio_array_push(q_dst, found);
	} else {
		// this is probably caused by work requesting from another core.
		// in that case, it's not a problem. It could also be that it is currently
		// executing, in which case we don't have to do anything.
	}

	core_unlock(task->cpu_no, lock);
}

void core_tq_update_blocked_task(enum lock_mode lm, task_t *task)
{
	int lock = core_lock(task->cpu_no, lm);
	core_t *core = core_get(task->cpu_no);

	for (int i = BLOCK_REASON_MIN; i <= BLOCK_REASON_MAX; i++) {

		prio_array_t *q = &(core->blocked_tasks[i]);

		// NOTE: this doesn't work with many blocked tasks (~500)
		// and pop_first + push. updating is obviously preferable,
		// but still: there might be a bug/race condition somewhere here...
		if (prio_array_update_first(q, &task_specific, task)) {
			break;
		}
	}

	core_unlock(task->cpu_no, lock);
}

task_t *core_tq_get_current()
{
	return core_get(ID)->current_task;
}

void core_tq_remove_current()
{
	assert(core_get(ID)->current_task != NULL);
	core_get(ID)->current_task = NULL;
}


unsigned int core_tq_size(unsigned int core_num, int sched_policy)
{
	if (core_num < *SYS_CPU_COUNT && (
		sched_policy == SCHED_OTHER ||
		sched_policy == SCHED_RR ||
		sched_policy == SCHED_FIFO))
	{
		return prio_array_size(&(core_get(core_num)->queue[sched_policy]));
	}
	return 0;
}


void core_tq_block_current(enum lock_mode lm, enum task_block_reason reason)
{
	core_t *me = core_get(ID);
	int lock = core_lock(ID, lm);

	if (me->current_task == NULL || !(me->current_task->blockable)) {
		return;
	}

	prio_array_push(&(me->blocked_tasks[reason]), me->current_task->se);

	me->current_task->state = BLOCKED;
	me->current_task = NULL;

	core_unlock(ID, lock);
}

void core_tq_unblock_tasks(enum lock_mode lm)
{
	static int (*should_pop[])(void*) = {
		[BLOCK_COND]		= &should_pop_cond_waiting,
		[BLOCK_COND_TIME]	= &should_pop_cond_time_waiting,
		[BLOCK_BARRIER]     = &should_pop_barrier_waiting
	};

	core_t *me = core_get(ID);
	int lock = core_lock(ID, lm);

	for (int r = BLOCK_REASON_MIN; r <= BLOCK_REASON_MAX; r++) {

		prio_array_t *q = &me->blocked_tasks[r];
		while (should_pop[r](prio_array_peek(q))) {

			sched_entity_t *se = prio_array_pop(q);
			task_t *task = se->task;

			task->state = RUNNING;
			task->cond_waiting = 0;
			task->cond_wait_timeout = 0;

			int pushed = prio_array_push(&(me->queue[task->attrs.schedpolicy]), se);
			if (!pushed) {
				linked_list_push(&(me->queued_tasks), se);
			}

		}
	}

	core_unlock(ID, lock);
}

int should_context_switch()
{
	core_t *me = core_get(ID);

	if (me->current_task == NULL) {
		return 1;
	}

	int sched_policy = SCHED_OTHER;
	if (me->active_queue == &me->queue[SCHED_FIFO]) {
		sched_policy = SCHED_FIFO;
	} else if (me->active_queue == &me->queue[SCHED_RR]) {
		sched_policy = SCHED_RR;
	}

	return (
			sched_policy == SCHED_OTHER && (
				prio_array_size(&me->queue[SCHED_FIFO]) > 0 ||
				prio_array_size(&me->queue[SCHED_RR]) > 0   || (
					prio_array_size(&me->queue[SCHED_OTHER]) > 0 &&
					!prio_array_beats_root(me->active_queue, me->current_task->se)
				)
			)
		) || (
			sched_policy == SCHED_RR && (
				prio_array_size(&me->queue[SCHED_FIFO]) > 0 || (
					prio_array_size(&me->queue[SCHED_RR]) > 0 &&
					!prio_array_beats_root(me->active_queue, me->current_task->se)
				)
			)
		) || (
			sched_policy == SCHED_FIFO && (
				prio_array_size(&me->queue[SCHED_FIFO]) > 0 &&
				!prio_array_beats_root(me->active_queue, me->current_task->se)
			)
		);
}

static void update_active_queue(core_t *core)
{
	if (prio_array_size(&(core->queue[SCHED_FIFO])) > 0) {
		core->active_queue = &core->queue[SCHED_FIFO];
	} else if (prio_array_size(&(core->queue[SCHED_RR])) > 0) {
		core->active_queue = &core->queue[SCHED_RR];
	} else {
		core->active_queue = &core->queue[SCHED_OTHER];
	}
}

static void push_residual_tasks()
{
	core_t *me = core_get(ID);

	linkedlist_item_t *head = linked_list_head(&(me->queued_tasks));
	while (head != NULL)
	{
		sched_entity_t *se = (sched_entity_t*)head->value;
		task_t *task = se->task;

		prio_array_t *q = &me->queue[task->attrs.schedpolicy];

		linkedlist_item_t *curr_item = head;
		head = head->next;

		if ((prio_array_capacity(q) - prio_array_size(q)) > 0) {

			sched_task_pre_push_hook(me, task);
			int pushed = prio_array_push(q, se);
			if (pushed) {
				task->state		= RUNNING;
				task->cpu_no	= ID;
				task->se->start	= *SYS_TICKS;

				linked_list_remove(&(me->queued_tasks), curr_item);
				free(curr_item);
			}
		}
	}
}

#if (defined(DEBUG) || defined(WARNINGS))
void pt(task_t *t)
{
	printf("t%d @ %x, %s. state %d", *t->thread, t->register_set.r[REG_PC],
		t->cond_waiting ? "waiting" : "not waiting", t->state);
}

void core_tq_print(unsigned int core_num)
{
	core_t *core = core_get(core_num);

	printf("===================================================\n");

	if (core->current_task != NULL)
	{
		printf("[%u] current task: ", core_num);
		pt(core->current_task);
		printf("\n");

	} else {
		printf("[%u] current task: NULL(!)\n", core_num);
	}

	prio_array_print(&core->queue[SCHED_OTHER]);
	printf("blocked tasks:\n");
	join_print(core_num);
	printf("\n======================================\n");
}
#endif
