#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "shmac/mutex.h"

#include "../debug.h"

#if defined(DEBUG) || defined(WARNINGS)
#include "../sched/sched.h"
#endif
#include "heap.h"

static int default_comparator(void *item1, void *item2)
{
    // default: don't care about order
    return 0;
}

#if (HEAP_ALLOW_GROW || HEAP_ALLOW_SHRINK)
static void heap_realloc(prio_array_t *array, unsigned int new_size)
{
    void *ptr = realloc(array->data, sizeof(void*) * new_size);
    if (ptr != NULL) {
        array->data = (void**)ptr;
        array->size = new_size;
    } else {
        warn("realloc failed!");
    }
}
#endif

static void siftdown(prio_array_t *array, unsigned int index)
{
    // the item to move down the heap
    unsigned int root = index;

    // while we're not off the end of the heap...
    while (2 * root + 1 < array->next_idx) {

        // first child
        unsigned int child = root * 2 + 1;

        // if there is a second child and it is greater than the first
        // child, we're going to swap with that one instead
        if (child < array->next_idx - 1 && array->cmp(
                array->data[child],
                array->data[child + 1]) < 0) {
            child++;
        }

        // if the largest child is greater than the current node, swap
        // and mark the new child as our current node for further
        // sift-downs
        if (array->cmp(array->data[root], array->data[child]) < 0) {
            void *tmp = array->data[root];
            array->data[root] = array->data[child];
            array->data[child] = tmp;

            root = child;
        } else {
            // no more swapping possible, we've reached the end
            return;
        }
    }
}

static void siftup(prio_array_t *array, unsigned int index)
{
    // while we haven't reached the root...
    while (index > 0) {

        // the parent to possibly swap with
        unsigned int parent = (index - 1) / 2;

        // swap if parent > child
        if (array->cmp(array->data[parent], array->data[index]) < 0) {
            void *tmp = array->data[parent];
            array->data[parent] = array->data[index];
            array->data[index] = tmp;

            index = parent;
        } else {
            // else return, we're done
            break;
        }
    }
}

void heap_init(prio_array_t *array)
{
    if (array->size == 0) {
        array->size = HEAP_DEFAULT_SIZE;
    }

    // memory for storage
    array->data = (void**) malloc(sizeof(void*) * (array->size));
    if (array->data == NULL) {
        array->size = 0;
        return;
    }

    // item comparison function
    if (array->cmp == NULL) {
        array->cmp = &default_comparator;
    }

    // point to first free spot in queue
    array->next_idx = 0;
}

int heap_push(prio_array_t *array, void *item)
{
#if HEAP_ALLOW_GROW
    if (array->next_idx >= array->size) {
        heap_realloc(array, array->size * 2);
    }
#endif

    if (array->next_idx >= array->size) {
        return 0;
    }

    // add to end, move up the heap
    array->data[array->next_idx] = item;

    // move end-pointer index
    array->next_idx++;

    siftup(array, array->next_idx - 1);

    return 1;
}

void *heap_pop(prio_array_t *array)
{
#if HEAP_ALLOW_SHRINK
    if (array->next_idx < array->size / 4) {
        heap_realloc(array, array->size / 2);
    }
#endif

    if (array->next_idx == 0) {
        return NULL;
    }

    // swap root with last item
    void *root = array->data[0];

    array->next_idx--;
    array->data[0] = array->data[array->next_idx];

    // move new root down the heap and return old root
    siftdown(array, 0);
    return root;
}

void *heap_pop_push(prio_array_t *array, void *item)
{
    // a combined pop/insert operation, which lets us use a single
    // sift-down instead of a sift-down _and_ a sift-up


    void *root = array->data[0];
    array->data[0] = item;

    siftdown(array, 0);
    return array->next_idx == 0 ? NULL : root;
}

int heap_pop_first(prio_array_t *array, predicate_t pred, void *prop, void **dst)
{
    for (int i = 0; i < array->next_idx; i++) {
        if (pred(array->data[i], prop)) {
            void *item = array->data[i];

            array->next_idx--;
            array->data[i] = array->data[array->next_idx];

            siftdown(array, i);
            siftup(array, i);

            *dst = item;
            return 1;
        }
    }

    return 0;
}

int heap_update_first(prio_array_t *array, predicate_t pred, void *prop)
{
    for (int i = 0; i < array->next_idx; i++) {
        if (pred(array->data[i], prop)) {
            siftdown(array, i);
            siftup(array, i);

            return 1;
        }
    }

    return 0;
}

void heap_each(prio_array_t *array, action_t action)
{
    for (int i = 0; i < array->next_idx; i++) {
        action(array->data[i]);
    }
}

void *heap_peek(prio_array_t *array)
{
    return (array->next_idx > 0)
        ? array->data[0]
        : NULL;
}

void heap_rebuild(prio_array_t *array)
{
    if (array->next_idx < 2) {
        return;
    }

    for (int i = (array->next_idx - 2) / 2; i >= 0; i--) {
        siftdown(array, i);
    }
}

int heap_beats_root(prio_array_t *array, void *item)
{
    // assumes heap is not empty

    return array->cmp(item, array->data[0]) >= 0;
}

size_t heap_size(prio_array_t *array)
{
    return array->next_idx;
}

size_t heap_capacity(prio_array_t *array)
{
    return array->size;
}


#if (defined(DEBUG) || defined(WARNINGS))
static void print_node(prio_array_t *t, unsigned int index, unsigned int indent)
{
    int ind = indent;
    while (ind--) {
        printf(" ");
    }

    unsigned int c1 = 2 * index + 1;
    unsigned int c2 = 2 * index + 2;

    struct sched_entity* se = (struct sched_entity*)t->data[index];
    task_t *ptr = se->task;

    printf("%d @ %x (sch = %d, pri = %d, ep = %d/%d, cld = %d/%d, wait = %d, state = %d)\n",
        *ptr->thread, ptr->register_set.r[REG_PC],
        ptr->attrs.schedpolicy, ptr->attrs.schedparam.sched_priority,
        ptr->epoch_count, ptr->se->epoch_sched_count,
        c1 < t->next_idx ? t->cmp(ptr, t->data[c1]) : 2,
        c2 < t->next_idx ? t->cmp(ptr, t->data[c2]) : 2,
        ptr->cond_waiting, ptr->state);

    if (c1 < t->next_idx) {
        print_node(t, c1, indent + 4);
        if (c2 < t->next_idx) {
            print_node(t, c2, indent + 4);
        }
    }
}
void heap_print(prio_array_t *array)
{
    printf("heap(count = %d, size = %d, addr = %x, def_size = %d, grows = %d, shrinks = %d)\n",
        array->next_idx, array->size, (unsigned int)array->data,
        HEAP_DEFAULT_SIZE, HEAP_ALLOW_GROW, HEAP_ALLOW_SHRINK);

    if (array->next_idx != 0)
        print_node(array, 0, 0);
}

int heap_validate(prio_array_t *array)
{
    for (int i = 0; i < array->next_idx; i++) {
        int c1 = 2 * i + 1;
        int c2 = 2 * i + 2;

        if (c1 < array->next_idx) {
            if (array->cmp(array->data[i], array->data[c1]) < 0) {
                return 0;
            }

            if (c2 < array->next_idx) {
                if (array->cmp(array->data[i], array->data[c2]) < 0) {
                    return 0;
                }
            }
        }

        for (int j = i + 1; j < array->next_idx; j++) {
            if (array->data[i] == array->data[j]) {
                warn("duplicate! found %d twice, at index %d and %d. max idx is %d, size %d",
                    (int)array->data[i], i, j, array->next_idx - 1, heap_size(array));
                return 0;
            }
        }
    }

    return 1;
}
#endif
