#ifndef _SHMAC_PTHREAD_HEAP_H_
#define _SHMAC_PTHREAD_HEAP_H_

/* 
 * a max-heap implementation of void* items.
 */

// this is the default size we allocate (times sizeof(void*)) if a size
// of zero is passed to heap_init.
#define HEAP_DEFAULT_SIZE	4

// should be grow the heap size if it is full? otherwise we block the operation
#define HEAP_ALLOW_GROW		1

// can the heap shrink again if the item count gets below a certain percentage
// of heap size? in our case that percentage is 25.
#define HEAP_ALLOW_SHRINK	0

typedef int (*predicate_t)(void *item, void *prop);
typedef void (*action_t)(void *item);

typedef struct {
    void                  **data;     // pointer to actual data
    size_t                  size;     // size of heap
    unsigned int            next_idx; // next free slot
    int (*cmp)(void*, void*);         // comparator for heap_items
} heap_t;

// initializes the heap, it that is necessary
void heap_init(heap_t *array);

// insert an item into the heap
int heap_push(heap_t *array, void *);

// removes the root element frmo the heap. will block
// until there is something to pop if the heap is empty
void *heap_pop(heap_t *array);

// A combined pop and push operation.
void *heap_pop_push(heap_t *array, void *item);

// removes the first item that matches predicate pred.
int heap_pop_first(heap_t *array, predicate_t pred, void *prop, void **dst);

// scans and updates (position in heap for) the first element that matches pred.
int heap_update_first(heap_t *array, predicate_t pred, void *prop);

// apply an action to all items in the heap.
void heap_each(heap_t *array, action_t action);

// returns the root element without removing it
void *heap_peek(heap_t *array);

// rebuilds the heap. useful for bulk updating
void heap_rebuild(heap_t *array);

// checks whether the provided item would have displaced
// the current root element it if were to be inserted.
int heap_beats_root(heap_t *array, void *item);

// gets the current element count if the heap.
size_t heap_size(heap_t *array);

// returns the capcity of the heap. This can change dynamically
// if HEAP_ALLOW_{GROW,SHRINK} are defined.
size_t heap_capacity(heap_t *array);

// used for debugging purposes: validates that a heap
// satifies the max-heap property and ensures that there are
// no duplicates (which are a bug in our case).
#if (defined(DEBUG) || defined(WARNINGS))
void heap_print(heap_t *array);
int heap_validate(heap_t *array);
#endif

#endif
