#include <stdlib.h>
#include <stdint.h>
#include "event_log.h"

#ifdef ENABLE_EVENT_LOGGING

#define MAX_VA_ARGS 5

struct event_struct {
	unsigned int timestamp;
	unsigned int core;
	enum log_event ev;

	task_t *task;
	pthread_t thread;

	unsigned int argc;
	unsigned int argv[MAX_VA_ARGS];
};

// the short names are so they fit nicely in a (gnu)plot :)
static char *strevents[] = {
	/* TASK_CREATED */ "creat",
	/* TASK_ASSIGNED */ "assign",
	/* TASK_EXECUTING */ "exec",
	/* TASK_STARTED */ "start",
	/* TASK_CONTINUED */ "cont",
	/* TASK_STOPPED */ "stop",
	/* TASK_CANCELED */ "canc",
	/* TASK_EXITED */ "exi",
	/* TASK_FINISHED */ "fini",

	/* joining */
	/* TASK_JOINING */ "jning",
	/* TASK_JOINED */ "jnd",
	/* TASK_JOIN_BLOCKED, */ "jnblk",
	/* TASK_JOIN_VIVIFIED, */ "jnviv",

	/* irq */
	/* IRQ_STARTED */ "irqs",
	/* IRQ_FAILED */ "irqerr",
	/* IRQ_FAILED_CORE_LOCK */ "irqerrc",
	/* IRQ_FAILED_MLOCK */ "irqerrm",
	/* IRQ_FINISHED */ "irqf",

	/* message queue-related */
	/* MQ_WORK_REQUEST_SENT */ "mqreqs",
	/* MQ_WORK_REQUEST_RECEIVED */ "mqreqr",
	/* MQ_WORK_RESPONSE_SENT */ "mqresps",
	/* MQ_WORK_RESPONSE_RECEIVED */ "mqrespr",

	/* global tq */
	/* GTQ_INIT */ "gtqinit",
	/* GTQ_INSERT */ "gtqins",
	/* GTQ_REMOVE */ "gtqrem",
	/* GTQ_LOOKUP_OK */ "gtqlutok",
	/* GTQ_LOOKUP_FAILED */ "gtqluterr"
};

#define HEADERS "timestamp,core,event_id,event_str,task_addr,task_id\n"

#define EVENT_BUFFER_SZ	8192

static struct event_struct *buf[2*MAX_CPUS];
static int buf_idx[2*MAX_CPUS];
static int logging_enabled[2*MAX_CPUS] = { [0 ... ((2 * MAX_CPUS) - 1)] = 1 };

void log_enable(unsigned int core)
{
	logging_enabled[core] = 1;
}
void log_disable(unsigned int core)
{
	logging_enabled[core] = 0;
}

static int mode()
{
	// to avoid deadlocking, we store event log for irq mode
	// and user mode separately so we dont have to lock.

	int x;
	asm("mrs %0, cpsr\n"
		: "=r"(x));
	return (x & 0x1f) != 0x10 ? 1 : 0;
}

void log_event(unsigned int core, enum log_event ev, task_t *t, int va_argc, ...)
{
	// copy SYS_TICKS here in case a significant amount of time
	// is spent waiting for the lock.
	unsigned int ts = *SYS_TICKS;
	int core_id = 2 * (*TILEREG_CPUID) + mode();

	if (!logging_enabled[core_id]) {
		return;
	}

	if (buf[core_id] == NULL) {
		buf[core_id] = (struct event_struct*)malloc(
				sizeof(struct event_struct) * EVENT_BUFFER_SZ);

		if (buf[core_id] == NULL) {
			printf("failed to alloc memory for logging!\n");
			logging_enabled[core_id] = 0;
		}

		debug("core %d logging to 0x%x, count @ 0x%x", core_id, buf[core_id], &buf_idx[core_id]);
		buf_idx[core_id] = 0;
	}

	int idx = buf_idx[core_id]++;
	if (idx >= EVENT_BUFFER_SZ) {
		log_disable(core_id);
		printf("event buffer full for %d!\n", core_id);
		return;
	}

	buf[core_id][idx].timestamp = ts;
	buf[core_id][idx].core = core;
	buf[core_id][idx].ev = ev;
	buf[core_id][idx].task = t;
	buf[core_id][idx].thread = (t == NULL) ? 0 : *t->thread;
	buf[core_id][idx].argc = va_argc;

	va_list valist;
	va_start(valist, va_argc);

	for (int i = 0; i < va_argc; i++) {
		buf[core_id][idx].argv[i] = va_arg(valist, int);
	}

	va_end(valist);
}

static void dump_log_internal(void (*event_func)(struct event_struct *ev))
{
	// disable all logs so we don't get any changes while we're dumping
	for (int i = 0; i < MAX_CPUS; i++) {
		log_disable(2 * i);
		log_disable(2 * i + 1);
	}

	//
	// find active logs -- this is just an optimization so we don't
	// scan all 2 * MAX_CPUS heads on every iteration. If we only
	// have 3 cores, we shouldn't have to scan more than 6 queues.
	//

	// the buffers there are non-empty; the ones we will loop over later
	struct event_struct *bufs[2 * MAX_CPUS];

	// we need the size of each one
	int len[2 * MAX_CPUS];

	// the number of buffers in 'bufs'
	int nof_bufs = 0;

	int nof_events = 0;

	for (int i = 0; i < 2 * MAX_CPUS; i++) {
		if (buf_idx[i] != 0) {
			bufs[nof_bufs] = buf[i];
			len[nof_bufs] = buf_idx[i];
			nof_bufs++;
			nof_events += buf_idx[i];
		}
	}

	printf("there are %d logs, %d events\n", nof_bufs, nof_events);

	//
	// now merge them
	//

	int curr_idx[nof_bufs];
	for (int i = 0; i < nof_bufs; i++) {
		curr_idx[i] = 0;
	}

	int ctr = 0;
	while (1) {
		int min_core = -1;
		int min_ts = INT32_MAX;

		// find the first event by comparing timestamps
		for (int core_idx = 0; core_idx < nof_bufs; core_idx++) {
			int head_idx = curr_idx[core_idx];

			if (head_idx < len[core_idx]) {
				if (bufs[core_idx][head_idx].timestamp < min_ts) {
					min_ts = bufs[core_idx][head_idx].timestamp;
					min_core = core_idx;
				}
			}
		}

		// break if we have exhausted all the non-empty event logs
		if (min_core == -1) break;

		int curr_id = curr_idx[min_core]++;
		event_func(&bufs[min_core][curr_id]);

		ctr++;
		if (ctr % 128 == 0) {
			printf("\r%d / %d      ", ctr, nof_events);
			fflush(stdout);
		}
	}

	printf("\ndone!\n");
}

void dump_log()
{
	void internal(struct event_struct *ev)
	{
		printf("%u,%u,%d,%s,0x%x,%u", ev->timestamp,
			ev->core, ev->ev, strevents[ev->ev],
			(unsigned int)ev->task, ev->thread);

		// print variable arguments
		for (int j = 0; j < ev->argc; j++) {
			printf(",%u", ev->argv[j]);
		}

		printf("\n");
	}

	printf(HEADERS);
	dump_log_internal(&internal);
}

void dump_log_mem(char **location, size_t *bytes_written)
{
	int size = 1024;
	int maxlen_item = 150;
	int final_padding = 10;

	char *mem = (char*)malloc(size);
	char *ptr = mem;
	char *end = mem + size;

	if (mem == NULL) {
		printf("failed to alloc memory!\n");
		return;
	}

	void internal(struct event_struct *ev)
	{
		// printf("%d = %s\n", ev->ev, strevents[ev->ev]);
		ptr += sprintf(ptr, "%u,%u,%d,%s,0x%x,%u",
					ev->timestamp, ev->core, ev->ev, strevents[ev->ev],
					(unsigned int)ev->task, ev->thread);

		// print variable arguments
		for (int j = 0; j < ev->argc; j++) {
			ptr += sprintf(ptr, ",%u", ev->argv[j]);
		}

		ptr += sprintf(ptr, "\n");

		if ((end - ptr) < maxlen_item) {
			int written = ptr - mem;
			size *= 2;
			mem = realloc(mem, size);

			if (mem == NULL) {
				printf("malloc failed! not dumping log to mem\n");
				return;
			}

			ptr = mem + written;
			end = mem + size;
		}
	}

	printf("dumping log to memory...\n");

	ptr += sprintf(mem, HEADERS);

	dump_log_internal(&internal);

	printf("\nDone!\n");

	// // in case the shmac_dump tool does any rounding we
	// // don't want it to: pad the output with some spaces.
	for (int i = 0; i < final_padding; i++)
		*ptr++ = ' ';

	*location = mem;
	*bytes_written = ptr - mem - final_padding;
}

#else
void log_enable(unsigned int core){}
void log_disable(unsigned int core){}
void dump_log(){}
void dump_log_mem(char **location, size_t *bytes_written)
{
	*location = NULL;
	*bytes_written = 0;
}
#endif
