#ifdef _POSIX_BARRIERS

#include <stdlib.h>
#include <pthread.h>
#include <assert.h>
#include <errno.h>

#include "../sched/sched.h"
#include "../debug.h"
#include "../core_tq.h"
#include "../data/rb_tree.h"
#include "../data/linked_list.h"


/*
 * Internals
 */

struct internal_barrier {
    // protects this struct and its waiters
    shmac_mutex_t mutex;

    // the user-supplied barrier
    pthread_barrier_t *barrier;

    // how many to wait for
    unsigned int count;

    // list of tasks already waiting here
    linkedlist_t waiters;
};


// Wake up every blocked task in the given list of waiters
static void notify_all(linkedlist_t *waiters)
{
    linkedlist_item_t *node = waiters->head;
    while (node != NULL) {
        ((task_t*)node->value)->barrier_waiting = 0;
        ((task_t*)node->value)->blockable = 1;

        node = node->next;
    }
}

// Free all malloced memory in 'waiters'
static void free_all(linkedlist_t *waiters)
{
    linkedlist_item_t *node = waiters->head;
    while (node != NULL) {
        linkedlist_item_t *next = node->next;

        linked_list_remove(waiters, node);
        free(node);

        node = next;
    }
}

// We keep a global tree of 'struct internal_barrier'
static rb_red_blk_tree *tree = NULL;
static shmac_mutex_t mlock = 0; // protects the tree

static int comp(const void *a, const void *b)
{
	return a < b ? (-1) : (a > b ? 1 : 0);
}

static void key_dest_func(void *_) {}
static void value_dest_func(void *_)  {}
static void print_func(const void *_) {}
static void print_info(void *_) {}
static int pthread_barrier_internal_init()
{
	if (tree == NULL) {
        lock_mutex(&mlock);
        if (tree == NULL) {
            tree = RBTreeCreate(
                comp,
                key_dest_func,
                value_dest_func,
                print_func,
                print_info);
        }

        unlock_mutex(&mlock);
		return tree != NULL;
	}

	return 1;
}

// Inserts a internal_barrier struct into the global rb-tree.
// initializes the list of threads waiting on the given barrier to be
// empty.
static int barrier_tree_insert(
        pthread_barrier_t *barrier,
        unsigned int count)
{
    pthread_barrier_internal_init();

    struct internal_barrier *internal = (struct internal_barrier*)
        malloc(sizeof(struct internal_barrier));

    if (internal == NULL) {
        return 0;
    }

    internal->mutex = SHMAC_MUTEX_INIT_VALUE;
    internal->barrier = barrier;
    internal->count = count;
    linked_list_init(&(internal->waiters));

    lock_mutex(&mlock);
    rb_red_blk_node *node = RBTreeInsert(tree, barrier, internal);
    unlock_mutex(&mlock);
    if (node == NULL) {
        return 0;
    }

    return 1;
}

// Remove the related internal_barrier from the global rb-tree. Notifies
// all tasks that were waiting that they should wake up.
static int barrier_tree_remove(
        pthread_barrier_t *barrier)
{
    lock_mutex(&mlock);
    rb_red_blk_node *node = RBExactQuery(tree, barrier);
    unlock_mutex(&mlock);
    if (node == NULL) {
        return 0;
    }

    struct internal_barrier *internal = (struct internal_barrier*)node->info;

    lock_mutex(&internal->mutex);

    notify_all(&internal->waiters);
    free_all(&internal->waiters);
    free(internal);

    unlock_mutex(&internal->mutex);

    return 1;
}

static int barrier_tree_wait(pthread_barrier_t *barrier)
{
    lock_mutex(&mlock);
    rb_red_blk_node *node = RBExactQuery(tree, barrier);
    unlock_mutex(&mlock);
    if (node == NULL) {
        return 0;
    }

    struct internal_barrier *internal = (struct internal_barrier*)node->info;
    lock_mutex(&internal->mutex);

    if (linked_list_size(&internal->waiters) == internal->count - 1) {

        notify_all(&internal->waiters);
        free_all(&internal->waiters);

        unlock_mutex(&internal->mutex);
    } else {
        task_t *me = core_tq_get_current();
        assert(me != NULL);

        me->blockable = 0;
        me->barrier_waiting = 1;

        linked_list_push(&internal->waiters, me);
        unlock_mutex(&internal->mutex);

        me->blockable = 1;

        schedule_try();
        while (me->barrier_waiting);
    }

    return 1;
}

/*
 * External API
 */

int pthread_barrier_init(pthread_barrier_t *restrict barrier,
       const pthread_barrierattr_t *restrict attr, unsigned count)
{
    if (barrier == NULL || count == 0) {
        errno = EINVAL;
        return EINVAL;
    }

    if (!barrier_tree_insert(barrier, count)) {
        errno = ENOMEM;
        return ENOMEM;
    }

    return 0;
}

int pthread_barrier_destroy(pthread_barrier_t *barrier)
{
    if (tree == NULL) {
        errno = EINVAL;
        return EINVAL;
    }

    if (!barrier_tree_remove(barrier)) {
        return EINVAL;
    }

    return 0;
}

int pthread_barrier_wait(pthread_barrier_t *barrier)
{
    if (tree == NULL) {
        errno = EINVAL;
        return EINVAL;
    }

    if (!barrier_tree_wait(barrier)) {
        return EINVAL;
    } else {
        return 0;
    }
}

#endif
