#include <pthread.h>
#include <errno.h>
#include <assert.h>

#include "shmac/shmac.h"
#include "shmac/swi.h"

#include "../core_tq.h"
#include "../data/linked_list.h"
#include "pthread_cond.h"

static int check_dynamic_initialization(pthread_cond_t *cond)
{
	if (cond != NULL && *cond == PTHREAD_COND_INITIALIZER) {
		static shmac_mutex_t dyn_init_cond = 0;
		lock_mutex(&dyn_init_cond);
		if (*cond == PTHREAD_COND_INITIALIZER) {
			if (pthread_cond_init(cond, NULL) != 0) {
				return 0;
			}
		}
		unlock_mutex(&dyn_init_cond);
	}

	return 1;
}
int pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
	if (cond == NULL || mutex == NULL) {
		errno = EINVAL;
		return EINVAL;
	}

	if (!check_dynamic_initialization(cond)) {
		return EINVAL;
	}

	pthread_cond_internal_t *internal = pthread_cond_internal_find(cond);
	if (internal == NULL) {
		errno = EINVAL;
		return EINVAL;
	}

	if (internal->mutex != NULL && internal->mutex != mutex) {
		errno = EINVAL;
		return EINVAL;
	}

	task_t *me = core_tq_get_current();
	assert(me != NULL);

	internal->mutex = mutex;
	linked_list_push(&(internal->listeners), (void*)me);

	// make sure interrupt doesnt deadlock us here
	me->blockable = 0;

	me->cond_waiting = 1;
	me->cond_wait_timeout = 0;

	// if an interrupt happens at this point, it would see that the
	// task could be blocked. However, since we are still holding a lock,
	// that could deadlock us. That's why the "blockable" flag is used
	// to prevent blocking.

	pthread_mutex_unlock(mutex);

	// _now_ we can move to blocked queue
	me->blockable = 1;

	schedule_try();

	while (me->cond_waiting);

	pthread_mutex_lock(mutex);

	return 0;
}

int pthread_cond_timedwait(pthread_cond_t *cond,
    pthread_mutex_t *mutex, const struct timespec *abstime)
{
	if (cond == NULL || mutex == NULL || abstime == NULL) {
		errno = EINVAL;
		return EINVAL;
	}

	if (!check_dynamic_initialization(cond)) {
		return ENOMEM;
	}

	unsigned int wakeup_time = abstime->tv_sec * SYS_TICKS_RATE +
		(unsigned int)(((long long)abstime->tv_nsec * SYS_TICKS_RATE) / (long long)1000000000);

	if (wakeup_time <= *SYS_TICKS) {
		pthread_mutex_lock(mutex);
		errno = ETIMEDOUT;
		return 0;
	}


	pthread_cond_internal_t *internal = pthread_cond_internal_find(cond);
	if (internal == NULL) {
		errno = EINVAL;
		return EINVAL;
	}


	if (internal->mutex != NULL && internal->mutex != mutex) {
		errno = EINVAL;
		return EINVAL;
	}

	task_t *me = core_tq_get_current();
	assert(me != NULL);

	internal->mutex = mutex;
	linked_list_push(&(internal->listeners), (void*)me);

	// make sure interrupt doesnt deadlock us here
	me->blockable = 0;

	me->cond_waiting = 1;
	me->cond_wait_timeout = wakeup_time;

	// if an interrupt happens at this point, it would see that the
	// task could be blocked. However, since we are still holding a lock,
	// that could deadlock us. That's why the "blockable" flag is used
	// to prevent blocking.

	pthread_mutex_unlock(mutex);

	// _now_ we can move to blocked queue
	me->blockable = 1;

	schedule_try();

	while (me->cond_waiting);

	pthread_mutex_lock(mutex);

	if (*SYS_TICKS > wakeup_time) {
		errno = ETIMEDOUT;
		return ETIMEDOUT;
	}

	return 0;
}
