#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>

#include "shmac/shmac.h"
#include "shmac/mutex.h"

#include "../core.h"
#include "../task.h"
#include "../core_tq.h"
#include "../global_tq.h"
#include "../event_log.h"
#include "../worker.h"


static shmac_mutex_t deadlock_mutex = SHMAC_MUTEX_INIT_VALUE;
static int will_deadlock(task_t *me, task_t *other)
{
    while (other != NULL) {
        if (other->waits_for == me) {
            return 1;
        }
        other = other->waits_for;
    }

    return 0;
}

static task_t *find_other_thread(pthread_t __pthread)
{
    int backoff = 1;
    int max_backoff = (1 << 15);

    pthread_t last_created;
    task_t *to_join = NULL;

    do {

        // try to find task from thread ID
        last_created = global_tq_last_created();

        if (__pthread <= last_created) {
            to_join = global_tq_find(__pthread);
        }

        // sleep to reduce lock contention in global_tq_find()
        if (to_join == NULL) {

            int n = backoff;
            while (n--);

            backoff <<= 1;
            if (backoff > max_backoff) {
                backoff = max_backoff;
            }
        }


    } while (to_join == NULL && last_created < __pthread);

    return to_join;
}

static void reap(task_t *my_task, task_t *to_join, void **__value_ptr)
{
    while (to_join->state != DEQUEUED);
    log_taskf(TASK_JOINED, my_task, 1, *to_join->thread);

    if (__value_ptr != NULL) {
        *__value_ptr = to_join->ret;
    }

    to_join->state = REAPED;
    free_task(to_join);
}

int pthread_join(pthread_t __pthread, void **__value_ptr)
{
    task_t *to_join = find_other_thread(__pthread);

    if (to_join == NULL) {
        errno = ESRCH;
        return ESRCH;
    }

    if (to_join->attrs.detachstate == PTHREAD_CREATE_DETACHED) {
        errno = ESRCH;
        return ESRCH;
    }

    task_t *me = core_tq_get_current();

    scheduler_pause();
    lock_mutex(&deadlock_mutex);

    if (will_deadlock(me, to_join)) {
        errno = EDEADLK;
        unlock_mutex(&deadlock_mutex);
        return EDEADLK;
    }

    me->waits_for = to_join;
    unlock_mutex(&deadlock_mutex);
    scheduler_resume();

    log_taskf(TASK_JOINING, me, 1, *to_join->thread);

    switch (to_join->state)
    {
        case QUEUED:
        case RUNNING:
        case BLOCKED:
        case FINISHED:
            // on the off-chance that we have a pending work request, we don't
            //want this task to be stealable. the receiving core doesn't check for
            // it and will just push it into its task heap, which we dont want.
            me->pinned = 1;

            // _now_ we can yield execution until the thread we're joining is finished
            me->join_waiting = 1;

            // speed up the process a bit :)
            schedule_try();

            // busy-wait
            reap(me, to_join, __value_ptr);
            break;
        case DEQUEUED:
            reap(me, to_join, __value_ptr);
            break;
        case REAPED:
            errno = EINVAL;
            return EINVAL;
        case CANCEL_PENDING:
        case CANCELED:
            errno = EINVAL;
            return EINVAL;
        default:
            warn("bug (unexpected state %d)!", to_join->state);
            break;
    }

    me->waits_for = NULL;

    return 0;
}
