#include <stdlib.h>
#include <pthread.h>
#include <errno.h>

#include "shmac/mutex.h"
#include "../data/rb_tree.h"

/*
 * Internals. We maintain thread-specific storage in a two-level RB-tree.
  * the first level maps thread ID to a thread-specific RB-tree. That RB-
  * tree maps from key to value (actually, to the key_value_pair struct)
 */

struct key_value_pair {
	pthread_key_t key;
	void *value;
	void (*destructor)(void*);
};

static rb_red_blk_tree *global_tree = NULL;
static int comp(const void *a, const void *b)
{
	return a < b ? (-1) : (a > b ? 1 : 0);
}

static void key_dest_func(void *_) {}
static void value_dest_func(void *_)  {}
static void print_func(const void *_) {}
static void print_info(void *_) {}

static shmac_mutex_t global_tree_mutex = SHMAC_MUTEX_INIT_VALUE;
static int global_tree_inited = 0;

static rb_red_blk_tree *pthread_thread_init(pthread_t tid)
{
	lock_mutex(&global_tree_mutex);

	// initialize global tree, which maintains a mapping from
	// thread ID to thread-specific RB-tree
	if (global_tree == NULL) {
		global_tree = RBTreeCreate(
			comp,
			key_dest_func,
			value_dest_func,
			print_func,
			print_info);

		global_tree_inited = global_tree != NULL;
	}

	// now initialize the tread-specific three if we have to
	if (global_tree_inited) {
		rb_red_blk_node *tid_node = RBExactQuery(global_tree, (void*)tid);
		if (tid_node == NULL) {
			rb_red_blk_tree *tid_tree = RBTreeCreate(
				comp,
				key_dest_func,
				value_dest_func,
				print_func,
				print_info);

			RBTreeInsert(global_tree, (void*)tid, (void*)tid_tree);
			unlock_mutex(&global_tree_mutex);

			return tid_tree;
		} else {
			unlock_mutex(&global_tree_mutex);
			return (rb_red_blk_tree*)(tid_node->info);
		}
	}

	unlock_mutex(&global_tree_mutex);
	return NULL;
}


/*
 * External API
 */


int pthread_key_create(pthread_key_t *key, void (*destructor)(void*))
{
	static shmac_mutex_t mutex = SHMAC_MUTEX_INIT_VALUE;
	static pthread_key_t unique_key = 0;

	pthread_t me = pthread_self();
	rb_red_blk_tree *tree = pthread_thread_init(me);

	if (tree == NULL) {
		errno = ENOMEM; // probably
		return ENOMEM;
	}

	lock_mutex(&mutex);
	*key = unique_key++;
	unlock_mutex(&mutex);

	struct key_value_pair *kvp = (struct key_value_pair*)malloc(sizeof(struct key_value_pair));
	if (kvp == NULL) {
		errno = ENOMEM;
		return ENOMEM;
	}

	kvp->key = *key;
	kvp->value = NULL;
	kvp->destructor = destructor;

	return RBTreeInsert(tree, (void*)*key, (void*)kvp) != NULL ? 0 : 1;
}

int pthread_key_delete(pthread_key_t key)
{
	rb_red_blk_tree *my_tree = pthread_thread_init(pthread_self());

	if (my_tree == NULL) {
		errno = ENOMEM;
		return ENOMEM;
	}

	rb_red_blk_node *node = RBExactQuery(my_tree, (void*)key);
	if (node == NULL) {
		errno = EINVAL;
		return EINVAL;
	}

	RBDelete(my_tree, node);
	return 0;
}

int pthread_setspecific(pthread_key_t key, const void *value)
{
	rb_red_blk_tree *my_tree = pthread_thread_init(pthread_self());

	if (my_tree == NULL) {
		errno = ENOMEM;
		return ENOMEM;
	}

	rb_red_blk_node *node = RBExactQuery(my_tree, (void*)key);
	if (node == NULL) {
		errno = EINVAL;
		return EINVAL;
	}

	struct key_value_pair *kvp = (struct key_value_pair*)node->info;
	kvp->value = (void*)value;

	return 0;
}
void *pthread_getspecific(pthread_key_t key)
{
	rb_red_blk_tree *my_tree = pthread_thread_init(pthread_self());

	if (my_tree == NULL) {
		errno = ENOMEM;
		return NULL;
	}

	rb_red_blk_node *node = RBExactQuery(my_tree, (void*)key);
	if (node == NULL) {
		errno = EINVAL;
		return NULL;
	}

	struct key_value_pair *kvp = (struct key_value_pair*)node->info;
	return kvp->value;
}

void pthread_keys_destroy()
{
	lock_mutex(&global_tree_mutex);

	// no thread has ever inserted a key, so no destructors to call
	if (global_tree == NULL) {
		unlock_mutex(&global_tree_mutex);
		return;
	}

	// _we_ havent inserted any keys, so no destructors to call
	rb_red_blk_node *global_node = RBExactQuery(global_tree, (void*)pthread_self());
	if (global_node == NULL) {
		unlock_mutex(&global_tree_mutex);
		return;
	}

	unlock_mutex(&global_tree_mutex);

	// walk through the tree and call any destructors we find
	rb_red_blk_tree *my_tree = (rb_red_blk_tree*)global_node->info;
	rb_red_blk_node *node = TreeFirstNode(my_tree);

	struct key_value_pair *kvp;

	while (node != my_tree->nil) {
		kvp = (struct key_value_pair*)node->info;

		if (kvp->value != NULL && kvp->destructor != NULL) {
			void *old_val = kvp->value;
			kvp->value = NULL;
			(kvp->destructor)(old_val);
		}

		node = TreeSuccessor(my_tree, node);
	}
}
