#include <stdlib.h>
#include <pthread.h>
#include <errno.h>

#include "../debug.h"
#include "../data/rb_tree.h"
#include "shmac/mutex.h"
#include "pthread_mutex.h"

/*
 * Internals
 */

static rb_red_blk_tree *tree = NULL;
static shmac_mutex_t mlock = 0;

static int comp(const void *a, const void *b)
{
	return a < b ? (-1) : (a > b ? 1 : 0);
}

static void key_dest_func(void *_) {}
static void value_dest_func(void *_)  {}
static void print_func(const void *_) {}
static void print_info(void *_) {}
static int pthread_mutex_tree_init()
{
	if (tree == NULL) {
		tree = RBTreeCreate(
			comp,
			key_dest_func,
			value_dest_func,
			print_func,
			print_info);

		return tree != NULL;
	}

	return 1;
}

pthread_mutex_internal_t *pthread_mutex_internal_find(pthread_mutex_t *mutex)
{
	if (!pthread_mutex_tree_init()) {
		return NULL;
	}

	rb_red_blk_node *node = RBExactQuery(tree, (void*)mutex);
	return node == NULL ? NULL : (pthread_mutex_internal_t*)node->info;
}

static int pthread_mutex_internal_insert(pthread_mutex_internal_t *mutex)
{
	lock_mutex(&mlock);
	if (!pthread_mutex_tree_init()) {
		unlock_mutex(&mlock);
		return 0;
	}

	pthread_mutex_internal_t *copy = (pthread_mutex_internal_t*)malloc(
		sizeof(pthread_mutex_internal_t));

	if (copy == NULL) {
		unlock_mutex(&mlock);
		return 0;
	}

	*copy = *mutex;
	RBTreeInsert(tree, (void*)(copy->mutex), (void*)copy);

	unlock_mutex(&mlock);
	return 1;
}

static int pthread_mutex_internal_remove(pthread_mutex_t *mutex)
{
	lock_mutex(&mlock);
	if (!pthread_mutex_tree_init()) {
		errno = ENOMEM;
		return ENOMEM;
	}

	rb_red_blk_node *node = RBExactQuery(tree, (void*)mutex);
	if (node != NULL) {
		RBDelete(tree, node);
		unlock_mutex(&mlock);
		return 0;
	}

	unlock_mutex(&mlock);

	errno = EINVAL;
	return EINVAL;
}


/*
 * External API
 */

int pthread_mutex_init(pthread_mutex_t *mutex,
    const pthread_mutexattr_t *attr)
{
	if (mutex == NULL) {
		errno = EINVAL;
		return EINVAL;
	}

	pthread_mutex_internal_t mutex_state = {
		.mutex = mutex,
		.owner = -1,
		.lock_count = 0
	};

	if (attr == NULL) {
		int ret;
		if ( (ret = pthread_mutexattr_init(&mutex_state.attr)) != 0) {
			return ret;
		}
	} else {
		mutex_state.attr = *attr;
	}

	if (!pthread_mutex_internal_insert(&mutex_state)) {
		*mutex = PTHREAD_MUTEX_INITIALIZER;
		errno = ENOMEM;
		return ENOMEM;
	}

	// this must be done after internal_insert because of the
	// dynamic mutex initialization done in pthread_mutex_lock.
	*mutex = SHMAC_MUTEX_INIT_VALUE;
	return 0;
}

int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
	pthread_mutex_internal_t *internal = pthread_mutex_internal_find(mutex);
	if (internal != NULL) {

		*mutex = PTHREAD_MUTEX_INITIALIZER;
		return pthread_mutex_internal_remove(mutex);
	}

	return 0;
}
