#include <pthread.h>
#include <errno.h>
#include <assert.h>

#include "../sched/sched.h"
#include "shmac/mutex.h"
#include "pthread_mutex.h"


static int check_dynamic_initialization(pthread_mutex_t *mutex)
{
	if (mutex != NULL && *mutex == PTHREAD_MUTEX_INITIALIZER) {
		static shmac_mutex_t dyn_init_mutex = 0;
		lock_mutex(&dyn_init_mutex);
		if (*mutex == PTHREAD_MUTEX_INITIALIZER) {
			if (pthread_mutex_init(mutex, NULL) != 0) {
				return 0;
			}
		}
		unlock_mutex(&dyn_init_mutex);
	}

	return 1;
}

// Lock
static int _internal_lock_normal(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	lock_mutex((volatile shmac_mutex_t*)internal->mutex);

	internal->owner = pthread_self();
	internal->lock_count = 1;

	debug("%d got mutex", internal->owner);
	return 0;
}
static int _internal_lock_errorcheck(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	if (internal->owner == pthread_self()) {
		errno = EDEADLK;
		return EDEADLK;
	}

	return _internal_lock_normal(internal);
}
static int _internal_lock_recursive(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	if (internal->owner == pthread_self()) {
		internal->lock_count++;
		return 0;
	}

	return _internal_lock_normal(internal);
}

// Unlock
static int _internal_unlock_normal(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	internal->owner = PTHREAD_MUTEX_OWNER_NOBODY;
	internal->lock_count = 0;

	unlock_mutex((volatile shmac_mutex_t*)internal->mutex);
	return 0;
}
static int _internal_unlock_errorcheck(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	if (internal->lock_count != 1 ||
		internal->owner != pthread_self())
	{
		debug("failed to unlock, count %d owner %d == %d?",
				internal->lock_count, internal->owner, pthread_self());
		errno = EPERM;
		return EPERM;
	}

	internal->owner = PTHREAD_MUTEX_OWNER_NOBODY;
	internal->lock_count = 0;

	unlock_mutex((volatile shmac_mutex_t*)internal->mutex);
	return 0;
}
static int _internal_unlock_recursive(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	if (internal->owner == pthread_self()) {
		if (internal->lock_count == 1) {
			return _internal_unlock_normal(internal);
		} else {
			internal->lock_count--;
			return 0;
		}

	} else {
		errno = EPERM;
		return EPERM;
	}
}

// Trylock
static int _internal_trylock_normal(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	int locked = trylock_mutex((volatile shmac_mutex_t*)internal->mutex);
	if (locked) {
		internal->owner = pthread_self();
		internal->lock_count = 1;

		return 0;
	} else {
		errno = EBUSY;
		return EBUSY;
	}
}
static int _internal_trylock_errorcheck(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	if (internal->owner == pthread_self()) {
		errno = EDEADLK;
		return EDEADLK;
	}

	return _internal_trylock_normal(internal);
}
static int _internal_trylock_recursive(pthread_mutex_internal_t *internal)
{
	assert(internal != NULL);

	if (internal->owner == pthread_self()) {
		internal->lock_count++;
		return 0;
	} else {
		int locked = trylock_mutex((volatile shmac_mutex_t*)internal->mutex);
		if (locked) {
			internal->owner = pthread_self();
			internal->lock_count = 1;

			return 0;
		} else {
			errno = EBUSY;
			return EBUSY;
		}
	}
}

//
// External API functions
//

int pthread_mutex_lock(pthread_mutex_t *mutex)
{
	if (!check_dynamic_initialization(mutex)) {
		return 1;
	}

	pthread_mutex_internal_t *internal = pthread_mutex_internal_find(mutex);

	if (internal != NULL) {

		debug("%d trying to get mutex", pthread_self());

		switch (internal->attr.type) {
#if defined(_UNIX98_THREAD_MUTEX_ATTRIBUTES)
			case PTHREAD_MUTEX_DEFAULT:
				return _internal_lock_normal(internal);
			case PTHREAD_MUTEX_ERRORCHECK:
				return _internal_lock_errorcheck(internal);
			case PTHREAD_MUTEX_RECURSIVE:
				return _internal_lock_recursive(internal);
#endif
			default:
				return _internal_lock_normal(internal);
		}

	} else {
		errno = EINVAL;
		return EINVAL;
	}
}

int pthread_mutex_trylock(pthread_mutex_t *mutex)
{
	if (!check_dynamic_initialization(mutex)) {
		return 1;
	}

	pthread_mutex_internal_t *internal = pthread_mutex_internal_find(mutex);

	if (internal != NULL) {

		switch (internal->attr.type) {
#if defined(_UNIX98_THREAD_MUTEX_ATTRIBUTES)
			case PTHREAD_MUTEX_DEFAULT:
				return _internal_trylock_normal(internal);
			case PTHREAD_MUTEX_ERRORCHECK:
				return _internal_trylock_errorcheck(internal);
			case PTHREAD_MUTEX_RECURSIVE:
				return _internal_trylock_recursive(internal);
#endif
			default:
				return _internal_trylock_normal(internal);
		}
	} else {
		errno = EINVAL;
		return EINVAL;
	}
}

int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
	pthread_mutex_internal_t *internal = pthread_mutex_internal_find(mutex);

	if (internal != NULL) {

		switch (internal->attr.type) {
#if defined(_UNIX98_THREAD_MUTEX_ATTRIBUTES)
			case PTHREAD_MUTEX_DEFAULT:
				return _internal_unlock_normal(internal);
			case PTHREAD_MUTEX_ERRORCHECK:
				return _internal_unlock_errorcheck(internal);
			case PTHREAD_MUTEX_RECURSIVE:
				return _internal_unlock_recursive(internal);
#endif
			default:
				return _internal_unlock_normal(internal);
		}

	} else {
		errno = EINVAL;
		return EINVAL;
	}
}
