#include <stdlib.h>

#include "shmac/mutex.h"

#include "sched/sched.h"
#include "timer.h"

#include "global_tq.h"
#include "event_log.h"

#define PIDHASH_SZ	16

struct locked_list {
	shmac_mutex_t lock;
	linkedlist_t linkedlist;
};

static struct locked_list pidhash[PIDHASH_SZ];
static volatile pthread_t last_created;
static shmac_mutex_t last_created_lock;

void global_tq_init()
{
	last_created = 0;
	last_created_lock = 0;

	for (int i = 0; i < PIDHASH_SZ; i++) {
		init_mutex(&pidhash[i].lock);
		linked_list_init(&pidhash[i].linkedlist);
	}

	log(GTQ_INIT);
}

void global_tq_insert(pthread_t key, task_t *value)
{

	unsigned int hash_key = key & (PIDHASH_SZ - 1);

	scheduler_pause();
	lock_mutex(&pidhash[hash_key].lock);
	linked_list_push(&pidhash[hash_key].linkedlist, value);
	unlock_mutex(&pidhash[hash_key].lock);
	scheduler_resume();

	lock_mutex(&last_created_lock);
	last_created = last_created > key ? last_created : key;
	unlock_mutex(&last_created_lock);

	log_task(GTQ_INSERT, value);
}

void global_tq_remove(pthread_t key)
{
	unsigned int hash_key = key & (PIDHASH_SZ - 1);
	scheduler_pause();
	lock_mutex(&pidhash[hash_key].lock);

	linkedlist_item_t *item = linked_list_head(&pidhash[hash_key].linkedlist);
	while (item != NULL) {
		if (*((task_t*)item->value)->thread == key) {
			log_task(GTQ_REMOVE, item->value);
			linked_list_remove(&pidhash[hash_key].linkedlist, item);
			free(item);
			break;
		}
		item = item->next;
	}

	unlock_mutex(&pidhash[hash_key].lock);
	scheduler_resume();
}

task_t *global_tq_find(pthread_t key)
{
	unsigned int hash_key = key & (PIDHASH_SZ - 1);
	scheduler_pause();
	lock_mutex(&pidhash[hash_key].lock);

	task_t *ptr = NULL;

	linkedlist_item_t *item = linked_list_head(&pidhash[hash_key].linkedlist);
	while (item != NULL) {
		task_t *t = (task_t*)(item->value);
		if (*t->thread == key) {
			ptr = t; break;
		}
		item = item->next;
	}

	unlock_mutex(&pidhash[hash_key].lock);
	scheduler_resume();

	if (ptr == NULL) {
		logf(GTQ_LOOKUP_FAILED, 1, key);
	} else {
		log_task(GTQ_LOOKUP_OK, ptr);
	}

	return ptr;
}

void global_tq_reset()
{
	for (int i = 0; i < PIDHASH_SZ; i++) {
		linkedlist_item_t *item = linked_list_head(&pidhash[i].linkedlist);
		while (item != NULL) {
			linkedlist_item_t *next = item->next;

			linked_list_remove(&pidhash[i].linkedlist, item);
			free(item);

			item = next;
		}
	}

	last_created = 0;
}

pthread_t global_tq_last_created()
{
	return last_created;
}

unsigned int global_tq_nof_tasks()
{
	int sum = 0;
	for (int i = 0; i < PIDHASH_SZ; i++) {
		sum += pidhash[i].linkedlist.count;
	}
	return sum;
}
