#include "shmac/shmac.h"

#include "../global_tq.h"
#include "../core.h"
#include "sched.h"
#include "heap.h"

#define se2t(ts,se) struct task_struct *(ts) = \
    (struct task_struct*)((struct sched_entity*)(se))->task


/* Comparison function for SCHED_OTHER */
static int task_comparator_heap_other(void *task1_se, void *task2_se)
{
    se2t(task1,task1_se);
    se2t(task2,task2_se);

    int s1 = task1->attrs.schedparam.sched_priority,
        s2 = task2->attrs.schedparam.sched_priority;

    // special case for idle function: always down-prioritize
    if (s1 == 0 && s2 != 0) return -1;
    if (s2 == 0 && s1 != 0) return  1;

    // if tasks have different priorities, order them by
    // 't->epoch_sched_count / t->priority' (in increasing order).

    int t1 = ((sched_entity_t*)task1_se)->epoch_sched_count * s2;
    int t2 = ((sched_entity_t*)task2_se)->epoch_sched_count * s1;

    return (t1 < t2) ? 1 : (t1 > t2 ? -1 : 0);
}


/* Comparison function for SCHED_RR */
static int task_comparator_heap_rr(void *task1_se, void *task2_se)
{
    // round-robin scheduling is simply ordering by
    // the number of times a task has been run.
    int t1 = ((sched_entity_t*)task1_se)->epoch_sched_count,
        t2 = ((sched_entity_t*)task2_se)->epoch_sched_count;

    return (t1 < t2) ? 1 : (t1 > t2 ? -1 : 0);
}

/* Comparison function for SCHED_FIFO */
static int task_comparator_heap_fifo(void *task1_se, void *task2_se)
{
    // order by (priority, created_time)
    se2t(task1,task1_se);
    se2t(task2,task2_se);

    int t1 = ((sched_entity_t*)task1_se)->created_time,
        t2 = ((sched_entity_t*)task2_se)->created_time;

    int s1 = task1->attrs.schedparam.sched_priority,
        s2 = task2->attrs.schedparam.sched_priority;

    if (s1 != s2)
        return (s1 > s2) ? 1 : (s1 < s2 ? -1 : 0);
    else
        return (t1 < t2) ? 1 : (t1 > t2 ? -1 : 0);
}


/*
 * Scheduler Initialization
 */

static int sched_heap_init_called = 0;
void sched_heap_init()
{
     int (*comps[])(void*,void*) = {
        &task_comparator_heap_other, // SCHED_OTHER = 0
        &task_comparator_heap_fifo,  // SCHED_FIFO = 1
        &task_comparator_heap_rr     // SCHED_RR = 2
    };

    // we need to disable stack/heap collision checking because
    // the threads have stacks allocated _on_ the heap. 
    extern int _disable_stack_heap_collision_check; // in syscalls/sbrk.c
    _disable_stack_heap_collision_check = 1;

    global_tq_init();

    // set all core states
    for (int i = 0; i < *SYS_CPU_COUNT; i++) {
        core_init(i, comps);
    }

    sched_heap_init_called = 1;
}

int sched_heap_inited()
{
    return sched_heap_init_called;
}


/*
 * Task hooks
 */

void sched_heap_task_start_hook(task_t *task)
{
	task->se->last_start = *SYS_TICKS;
	log_task(TASK_STARTED, task);
}
void sched_heap_task_retire_hook(task_t *task)
{
	log_task(TASK_STOPPED, task);
}
void sched_heap_task_continue_hook(task_t *task)
{
	task->se->last_start = *SYS_TICKS;
	log_task(TASK_CONTINUED, task);
}
void sched_heap_task_tick_hook(task_t *task)
{
	uint32_t ticks = *SYS_TICKS;

	task->epoch_count++;
	task->se->epoch_sched_count++;

	task->se->vruntime += ticks - task->se->last_start;
}

void sched_heap_task_pre_push_hook(core_t* core, task_t *task)
{
    // 
    // Equalize priority at this point.
    //

	if (core->current_task != NULL && core->current_task != task) {

		// we want (new->epoch / new->prio) == (old->epoch / old->prio)

		task->se->epoch_sched_count = (core->current_task->se->epoch_sched_count *
				task->attrs.schedparam.sched_priority) /
			core->current_task->attrs.schedparam.sched_priority;
	// } else {
		// this is either an idle task or the main task before the
		// idle task on core 0 has started. or a bug.
	}
}


void sched_heap_entity_init(struct sched_entity *se, struct task_struct *task)
{
    se->task = task;
    task->se = se;

    se->epoch_sched_count = 0;

    se->vruntime          = 0;
    se->created_time      = *SYS_TICKS;
    se->last_start        = 0;
    se->start             = 0;
}
