#ifndef SHMAC_PTHREADS_SCHED_H_
#define SHMAC_PTHREADS_SCHED_H_

#include <pthread.h>
#include <limits.h>

#include "shmac/mutex.h"
#include "shmac/swi.h"

#include "enums.h"
#include "task.h"

#include "../debug.h"
#include "../event_log.h"

#include "../data/ring_buffer.h"
#include "../data/linked_list.h"

// Before a task is passed to a specific CPU, it is given this number. It will never
// again be unassigned once it is assigned, even if it is "in transit" to another core
// for load balancing purposes.
#define CPU_NO_UNASSIGNED   -1

// The special pthread_t ID used to designate the main thread of execution,
// which is not actually a thread.
#define PTHREAD_SELF_MASTER UINT_MAX    // a dummy thread id for core 0

#define PTHREAD_STACK_MIN   131072 // 128k
#define NOF_SCHED_CLASSES   3

#define schedule_try()     swi_trigger(SWI_PTHREAD_TRY_SCHEDULE)
#define scheduler_pause()
#define scheduler_resume()

#define SCHED_HEAP 0
#define SCHED_LIST 1 // not yet implemented

// the backend to use for the scheduler
#define CURRENT_SCHED_IMPL SCHED_HEAP

#if CURRENT_SCHED_IMPL == SCHED_HEAP
#include "../data/heap.h"
typedef heap_t prio_array_t;

#elif CURRENT_SCHED_IMPL == SCHED_LIST
#include "../data/linked_list.h"
typedef linkedlist_t prio_array_t;

#endif // if CURRENT_SCHED_NAME == ...

// the structure of messages that can be passed between cores.
typedef struct {
    enum core_msg   message;
    void           *extra;
} core_msg_t;

// encapsulates all the needed core-local queues and buffers.
struct core_struct {
    enum core_state state;                  // core state, as set by the core itself

    // Blocked tasks
    linkedlist_t    queued_tasks;           // singly-linked list of queued tasks
    linkedlist_t    joined_tasks;           // tasks that are blocked waiting for others

    // MQ
    int             has_queued_msgs;        // used to avoid scanning all MAX_CPUS ring buffers
    ring_buffer_t   queued_msgs[MAX_CPUS];  // fifo inter-core message queue,
                                            // one for each core to avoid locking

    // Active tasks
    prio_array_t    queue[NOF_SCHED_CLASSES]; // the three queues used: other/fifo/rr
    prio_array_t   *active_queue;             // a pointer to one of the fifo/rr/other queues
    task_t         *current_task;             // current task, not found in a task queue

    // Blocked in priority queues
    prio_array_t    blocked_tasks[BLOCK_REASON_MAX+1];// tasks that are waiting for some condition.

    // Misc.
    unsigned int    epoch_count;            // total # of time slices run on this core
    shmac_mutex_t   lock_object;            // lock for all access to this core_t
};
typedef struct core_struct core_t;


// This if/else thing is split because we want core_t to be defined
// here. Thus, we typedef prio_array_t and include the relevant header
// file _above the core_t declaration, but include this below it.

#if CURRENT_SCHED_IMPL == SCHED_HEAP
#include "heap.h"

#elif CURRENT_SCHED_IMPL == SCHED_LIST
#include "list.h"

#endif // if CURRENT_SCHED_NAME == ...

#endif // H guard
