#include <stdlib.h>

#include "shmac/shmac.h"
#include "data/linked_list.h"

#include "core_tq.h"
#include "sched/sched.h"
#include "task.h"

static void set_sched_policy(pthread_attr_t *att)
{
    if (att->inheritsched == PTHREAD_INHERIT_SCHED) {
        task_t *calling_thread = core_tq_get_current();

        if (calling_thread == NULL) {
            att->schedpolicy = SCHED_OTHER;
        } else {
            att->schedpolicy = calling_thread->attrs.schedpolicy;
        }
    }
}

static int init_or_clone_attributes(pthread_attr_t *dst, const pthread_attr_t *src)
{
    if (src == NULL) {
        dst->stacksize = 0;
        dst->stackaddr = NULL;

        if (pthread_attr_init(dst) != 0) {
            return 0;
        }
    } else {
        *dst = *src;
    }

    set_sched_policy(dst);
    return 1;
}

sched_entity_t *task_create(pthread_t *t, const pthread_attr_t *attrs,
                   void *(*func)(void *), void *arg,
                   int stack_free, int task_free, int pinned,
                   int cancel_state)
{
	static shmac_mutex_t tid_mutex  = SHMAC_MUTEX_INIT_VALUE;

    // the first MAX_CPUS IDs are reserved for idle tasks
    static pthread_t thread_id      = MAX_CPUS;

    pthread_attr_t att;
    if (!init_or_clone_attributes(&att, attrs)) {
        return NULL;
    }

    task_t *task = (task_t*)malloc(sizeof(task_t));
    if (task == NULL) {
        return NULL;
    }

    sched_entity_t *se = (sched_entity_t*)malloc(sizeof(sched_entity_t));
    if (se == NULL) {
        free(task);
        return NULL;
    }

    task->state      = QUEUED;
    task->cpu_no     = CPU_NO_UNASSIGNED;

    task->thread     = t;
    task->attrs      = att;

    task->stack_free = stack_free;
    task->task_free  = task_free;

    task->task       = func;
    task->arg        = arg;
    task->ret        = NULL;

    task->waits_for          = NULL;
    task->join_ll_item.value = NULL;
    task->join_ll_item.prev  = NULL;
    task->join_ll_item.next  = NULL;

    task->pinned        = pinned;
    task->blockable     = 1;

    task->cond_waiting      = 0;
    task->cond_wait_timeout = 0;
    task->join_waiting      = 0;
    task->barrier_waiting   = 0;

    task->cancel_type  = PTHREAD_CANCEL_ASYNCHRONOUS;
    task->cancel_state = cancel_state;

    task->epoch_count   = 0;

    for (int reg = 0; reg < ARM_REG_COUNT; reg++) {
        task->register_set.r[reg] = 0;
    }
    task->register_set.r[REG_SPSR] = 0x10;
    task->register_set.r[REG_R0] = (int)func;

    linked_list_init(&(task->cleanup_handlers));

    lock_mutex(&tid_mutex);
    *t = thread_id++;
    unlock_mutex(&tid_mutex);

    sched_entity_init(se, task);

    return se;
}
