#include <assert.h>
#include <stdio.h>

#include "../sched/sched.h"
#include "../data/heap.h"

#include "test.h"
#include "heap.h"

static prio_array_t heap;

static int comp(void *a, void *b)
{
	int x = (int)((task_t*)a)->thread;
	int y = (int)((task_t*)b)->thread;

	return (x < y) ? (-1) : (x > y ? 1 : 0);
}

static void test_heap_init()
{
	heap.data = NULL;
	heap.size = 0;
	heap.next_idx = 0;
	heap.cmp = &comp;

	heap_init(&heap);
}

int test_heap_all()
{
	test_heap_empty();
	test_heap_size();

	test_heap_peek();
	test_heap_beats_root();

	test_heap_push();
	test_heap_pop();
	test_heap_pop_push();

	return 1;
}

void test_heap_empty()
{
	test_heap_init();
	print_function_name();

	assert(heap_size(&heap) == 0);
	assert(heap_capacity(&heap) > 0);
	assert(heap_peek(&heap) == NULL);

	// pop blocks, so cannot test it
}

void test_heap_size()
{
	test_heap_init();
	print_function_name();

	task_t t[] = {
		{ .thread = (pthread_t*)1 },
		{ .thread = (pthread_t*)2 },
		{ .thread = (pthread_t*)3 },
		{ .thread = (pthread_t*)4 },
		{ .thread = (pthread_t*)5 },
		{ .thread = (pthread_t*)6 },
		{ .thread = (pthread_t*)7 },
		{ .thread = (pthread_t*)8 }
	};

	int n = sizeof(t) / sizeof(t[0]);
	int times = HEAP_DEFAULT_SIZE * 8;
	int pushed = 0;

	while (times--) {
		for (int i = 0; i < n; i++) {
			heap_push(&heap, &t[i]);
			assert(heap_size(&heap) == ++pushed);

		}
	}
}

void test_heap_push()
{
	test_heap_init();
	print_function_name();

	pthread_t tids[] = { 1,2,3,4,5,6,7,8 };

	task_t t[sizeof(tids)/sizeof(tids[0])];
	for (int i = 0; i < sizeof(tids)/sizeof(tids[0]); i++) {
		t[i].thread = &tids[i];
	}

	int n = sizeof(t) / sizeof(t[0]);
	int pushed = 0;
	int max = 0;

	for (int i = 0; i < n; i++) {
		max = max > (int)t[i].thread ? max : (int)*t[i].thread;

		int pushed_ok = heap_push(&heap, &t[i]);

		// push shouldn't fail (assuming heap can grow)
		assert(pushed_ok == 1);

		pushed++;

		// root should be highest pushed
		assert(*((task_t*)heap_peek(&heap))->thread == max);

		// heap has expected # of items
		assert(heap_size(&heap) == pushed);

		// capacity is increased as we push items
		assert(heap_capacity(&heap) >= heap_size(&heap));

		assert(heap_validate(&heap) == 1);
	}
}

void test_heap_pop()
{
	test_heap_init();
	print_function_name();

	task_t t[] = {
		{ .thread = (pthread_t*)1 },
		{ .thread = (pthread_t*)8 },
		{ .thread = (pthread_t*)5 },
		{ .thread = (pthread_t*)7 },
		{ .thread = (pthread_t*)3 },
		{ .thread = (pthread_t*)6 },
		{ .thread = (pthread_t*)4 },
		{ .thread = (pthread_t*)2 }
	};

	int n = sizeof(t) / sizeof(t[0]);
	int items = 0;

	for (int i = 0; i < n; i++) {
		heap_push(&heap, &t[i]);
		items++;
	}

	assert(heap_validate(&heap) == 1);

	// pop, and check that items are in order
	task_t *last = (task_t*)heap_pop(&heap);
	for (int i = 1; i < n; i++) {
		task_t *item = (task_t*)heap_pop(&heap);

		assert(last->thread >= item->thread);
		item = last;
	}

	assert(heap_validate(&heap) == 1);
}

void test_heap_pop_push()
{
	test_heap_init();
	print_function_name();

	task_t t[] = {
		{ .thread = (pthread_t*)1 },
		{ .thread = (pthread_t*)8 },
		{ .thread = (pthread_t*)5 },
		{ .thread = (pthread_t*)7 },
		{ .thread = (pthread_t*)3 },
		{ .thread = (pthread_t*)6 },
		{ .thread = (pthread_t*)4 },
		{ .thread = (pthread_t*)2 }
	};

	// push items and verify expected state
	heap_push(&heap, &t[2]);
	heap_push(&heap, &t[6]);
	heap_push(&heap, &t[7]);
	assert(heap_size(&heap) == 3);
	assert(((task_t*)heap_peek(&heap))->thread == t[2].thread);
	assert(heap_validate(&heap) == 1);

	// swap with smaller item and ensure it is not root
	task_t *old_root = (task_t*)heap_pop_push(&heap, &t[4]);
	assert(old_root->thread == t[2].thread);
	assert(((task_t*)heap_peek(&heap))->thread != t[4].thread);
	assert(heap_validate(&heap) == 1);

	// push larger item and ensure it is root
	old_root = (task_t*) heap_pop_push(&heap, &t[1]); 
	assert(old_root->thread == t[6].thread);
	assert(((task_t*)heap_peek(&heap))->thread == t[1].thread);
	assert(heap_validate(&heap) == 1);

	assert(heap_size(&heap) == 3);
}

void test_heap_peek()
{
	test_heap_init();
	print_function_name();	

	task_t t[] = {
		{ .thread = (pthread_t*)1 },
		{ .thread = (pthread_t*)8 },
		{ .thread = (pthread_t*)5 },
		{ .thread = (pthread_t*)7 },
		{ .thread = (pthread_t*)3 },
		{ .thread = (pthread_t*)6 },
		{ .thread = (pthread_t*)4 },
		{ .thread = (pthread_t*)2 }
	};

	heap_push(&heap, &t[0]);
	assert(heap_peek(&heap) == &t[0]);

	heap_push(&heap, &t[1]);
	assert(heap_peek(&heap) == &t[1]);

	heap_push(&heap, &t[2]);
	assert(heap_peek(&heap) == &t[1]);

	heap_push(&heap, &t[3]);
	assert(heap_peek(&heap) == &t[1]);

	heap_push(&heap, &t[4]);
	assert(heap_peek(&heap) == &t[1]);

	heap_push(&heap, &t[5]);
	assert(heap_peek(&heap) == &t[1]);

	heap_push(&heap, &t[6]);
	assert(heap_peek(&heap) == &t[1]);

	heap_push(&heap, &t[7]);
	assert(heap_peek(&heap) == &t[1]);

	assert(heap_validate(&heap) == 1);
}

void test_heap_beats_root()
{
	test_heap_init();
	print_function_name();

	task_t t[] = {
		{ .thread = (pthread_t*)1 },
		{ .thread = (pthread_t*)8 },
		{ .thread = (pthread_t*)5 },
		{ .thread = (pthread_t*)7 },
		{ .thread = (pthread_t*)3 },
		{ .thread = (pthread_t*)6 },
		{ .thread = (pthread_t*)4 },
		{ .thread = (pthread_t*)2 }
	};

	int n = sizeof(t) / sizeof(t[0]);
	for (int i = 0; i < n; i++) {
		heap_push(&heap, &t[i]);
	}

	task_t t1  = { .thread = (pthread_t*)100 },
		   t2 =  { .thread = (pthread_t*) 0 };

   void *beats  = (void*)&t1,
   		*doesnt = (void*)&t2;

	assert(heap_beats_root(&heap, beats) == 1);
	assert(heap_beats_root(&heap, doesnt) == 0);
}
