#include <stdio.h>
#include <assert.h>

#include  "../data/linked_list.h"

#include "test.h"
#include "linked_list.h"


static linkedlist_t list;

static void test_init()
{
	linked_list_init(&list);
}

static void pl(linkedlist_t *list)
{
	if (list == NULL) {
		printf("(empty)\n");
		return;
	}

	linkedlist_item_t *item = list->head;

	if (item != NULL) { 
		printf("h(%d) ", (int)item->value);
	}

	while (item != NULL) {
		printf("%d[p=%d,n=%d]", (int)item->value,
				item->prev == NULL ? 0 : (int)item->prev->value,
				item->next == NULL ? 0 : (int)item->next->value);

		if (item->next != NULL) {
			printf(" <-> ");
		}

		item = item->next;
	}

	if (linked_list_tail(list) != NULL) {
		printf(" t(%d) %s (size %u)", (int)linked_list_tail(list)->value,
			linked_list_head(list) == linked_list_tail(list)
				? "head==tail" : "head!=tail",
			linked_list_size(list));
	}

	printf("\n");
}

int test_linkedlist_all()
{
	test_linkedlist_empty();
	test_linkedlist_pop();
	test_linkedlist_push();
	test_linkedlist_remove();
	test_linkedlist_remove2();
	test_linkedlist_size();

	return 1;
}

void test_linkedlist_empty()
{
	test_init();
	print_function_name();

	assert(list.head == NULL);
	assert(list.tail == NULL);

	assert(linked_list_head(&list) == NULL);
	assert(linked_list_tail(&list) == NULL);

	assert(list.count == 0);
	assert(linked_list_size(&list) == 0);
}

void test_linkedlist_pop()
{
	test_init();
	print_function_name();

	linked_list_push(&list, (void*)1);
	linked_list_push(&list, (void*)2);
	linked_list_push(&list, (void*)3);

	assert(linked_list_tail(&list)->value == (void*)3);

	void *p1 = linked_list_pop(&list)->value;
	void *p2 = linked_list_pop(&list)->value;
	void *p3 = linked_list_pop(&list)->value;

	assert(linked_list_tail(&list) == NULL);

	assert(p1 == (void*)1);
	assert(p2 == (void*)2);
	assert(p3 == (void*)3);
}

void test_linkedlist_push()
{
	test_init();
	print_function_name();

	linked_list_push(&list, (void*)1);

	assert(linked_list_head(&list) != NULL);
	assert(linked_list_tail(&list) != NULL);

	assert(linked_list_head(&list)->value == (void*)1);
	assert(linked_list_tail(&list)->value == (void*)1);

	linked_list_push(&list, (void*)2);
	assert(linked_list_head(&list)->value == (void*)1);
	assert(linked_list_tail(&list)->value == (void*)2);

	linked_list_push(&list, (void*)3);
	assert(linked_list_head(&list)->value == (void*)1);
	assert(linked_list_tail(&list)->value == (void*)3);
}

void test_linkedlist_remove()
{
	test_init();
	print_function_name();

	// 1 <-> 2 <-> 3 <-> 4
	linked_list_push(&list, (void*)1);
	linked_list_push(&list, (void*)2);
	linked_list_push(&list, (void*)3);
	linked_list_push(&list, (void*)4);


	linkedlist_item_t *head = linked_list_head(&list);
	linkedlist_item_t *tail = linked_list_tail(&list);

	// check head/tail is set correctly
	assert(head != NULL);
	assert(tail != NULL);

	// correct chaining
	assert(head->value == (void*)1);
	assert(head->next->value == (void*)2);
	assert(head->next->next->value == (void*)3);
	assert(head->next->next->next->value == (void*)4);
	assert(head->next->next->next->next == NULL);

	// correct size
	assert(linked_list_size(&list) == 4);

	// now start removing the second item...
	linkedlist_item_t *two = head->next;
	linked_list_remove(&list, two);
	assert(head->value == (void*)1);
	assert(head->next->value == (void*)3);
	assert(linked_list_size(&list) == 3);

	// remove second element again
	linkedlist_item_t *three = head->next;
	linked_list_remove(&list, three);
	assert(head->value == (void*)1);
	assert(head->next->value == (void*)4);
	assert(linked_list_size(&list) == 2);

	// ... and again
	linkedlist_item_t *four = head->next;
	linked_list_remove(&list, four);
	assert(head->next == NULL);
	assert(linked_list_size(&list) == 1);

	// now we should only have the first element (1) left
	assert(linked_list_head(&list) != NULL);
	assert(linked_list_head(&list)->value == (void*)1);
	assert(linked_list_tail(&list) != NULL);
	assert(linked_list_tail(&list)->value == (void*)1);

	// now remove the last element
	linked_list_remove(&list, list.head);

	// ... and check that everything is reset to NULL
	assert(linked_list_size(&list) == 0);
	assert(linked_list_head(&list) == NULL);
	assert(linked_list_tail(&list) == NULL);
}

void test_linkedlist_remove2()
{
	test_init();
	print_function_name();

	// 1 <-> 2 <-> 3 <-> 4 <-> 5
	linked_list_push(&list, (void*)1);
	linked_list_push(&list, (void*)2);
	linked_list_push(&list, (void*)3);
	linked_list_push(&list, (void*)4);
	linked_list_push(&list, (void*)5);

	linkedlist_item_t *head = linked_list_head(&list);
	linkedlist_item_t *tail = linked_list_tail(&list);

	// new state: 2 <-> 3 <-> 4 <-> 5
	linked_list_remove(&list, head);
	assert((int)(linked_list_head(&list)->value) == 2);

	// doesnt exist!
	linked_list_remove(&list, head);
	assert((int)(linked_list_head(&list)->value) == 2);
	assert(linked_list_size(&list) == 4);

	// new state: 2 <-> 4 <-> 5
	linked_list_remove(&list, linked_list_head(&list)->next);
	assert(linked_list_head(&list)->value == (void*)2);
	assert(linked_list_head(&list)->next->value == (void*)4);
	assert(linked_list_head(&list)->next->next->value == (void*)5);
	assert(linked_list_head(&list)->next->prev->value == (void*)2);

	// new state: 2 <-> 4
	linked_list_remove(&list, tail);
	assert(linked_list_head(&list)->value == (void*)2);
	assert(linked_list_head(&list)->next->value == (void*)4);
	assert(linked_list_tail(&list)->value == (void*)4);
	assert(linked_list_tail(&list)->prev->value == (void*)2);


}

void test_linkedlist_size()
{
	test_init();
	print_function_name();

	assert(linked_list_size(&list) == 0);

	linked_list_push(&list, (void*)1);
	assert(linked_list_size(&list) == 1);

	linked_list_push(&list, (void*)1);
	assert(linked_list_size(&list) == 2);

	linked_list_push(&list, (void*)1);
	assert(linked_list_size(&list) == 3);

	linked_list_pop(&list);
	assert(linked_list_size(&list) == 2);

	linked_list_pop(&list);
	assert(linked_list_size(&list) == 1);

	linked_list_pop(&list);
	assert(linked_list_size(&list) == 0);
}
