#include <stdio.h>
#include <assert.h>
#include <pthread.h>

#include "shmac/mutex.h"

#include "test.h"
#include "pthreads_misc.h"


static void test_init()
{

}

int test_pthreads_misc_all()
{
	test_pthreads_misc_once();
	test_pthreads_misc_cleanup();

	test_pthreads_misc_keys();
	test_pthreads_misc_key_destructors();

	return 1;
}


/* Tests that pthread_once ensures that the specified routine
 * is only called once. Tested from both from main thread of
 * execution and separate thread.
 */
static int once_called_ctr = 0;
static void once_routine(void)
{
	static shmac_mutex_t mutex = SHMAC_MUTEX_INIT_VALUE;
	lock_mutex(&mutex);
	once_called_ctr++;
	unlock_mutex(&mutex);
}
static void *once_caller_thread(void *control)
{
	pthread_once((pthread_once_t*)control, &once_routine);
	return NULL;
}
void test_pthreads_misc_once()
{
	test_init();
	print_function_name();

	pthread_once_t control = PTHREAD_ONCE_INIT;
	pthread_once(&control, &once_routine);

	pthread_t tids[3];

	int c1_ok = pthread_create(&tids[0], NULL, &once_caller_thread, &control);
	int c2_ok = pthread_create(&tids[1], NULL, &once_caller_thread, &control);
	int c3_ok = pthread_create(&tids[2], NULL, &once_caller_thread, &control);

	int j1_ok = pthread_join(tids[0], NULL);
	int j2_ok = pthread_join(tids[1], NULL);
	int j3_ok = pthread_join(tids[2], NULL);

	assert(c1_ok == 0);
	assert(c2_ok == 0);
	assert(c3_ok == 0);

	assert(j1_ok == 0);
	assert(j2_ok == 0);
	assert(j3_ok == 0);

	assert(once_called_ctr == 1);
}

/* Tests that cleanup routines are called in the reverse order
 * of what they were specified in.
 */
static int cleanup_index = 0;
static void *cleanup_values[3];
static void cleanup_routine(void *arg)
{
	cleanup_values[cleanup_index++] = arg;
}
static void *cleanup_thread_func(void *arg)
{
	pthread_cleanup_push(&cleanup_routine, (void*)1);
	pthread_cleanup_push(&cleanup_routine, (void*)2);
	pthread_cleanup_push(&cleanup_routine, (void*)3);
	return NULL;
}
void test_pthreads_misc_cleanup()
{
	test_init();
	print_function_name();

	pthread_t tid;
	pthread_create(&tid, NULL, &cleanup_thread_func, NULL);
	pthread_join(tid, NULL);

	// ensure they were called in reverse order
	assert(cleanup_values[0] == (void*)3);
	assert(cleanup_values[1] == (void*)2);
	assert(cleanup_values[2] == (void*)1);
}

void test_pthreads_misc_keys()
{
	test_init();
	print_function_name();

	void *thread(void *arg)
	{
		pthread_key_t k1, k2;
		pthread_key_create(&k1, NULL);
		pthread_key_create(&k2, NULL);

		pthread_setspecific(k1, (void*)42);
		pthread_setspecific(k1, (void*)24);
		pthread_setspecific(k2, (void*)50);

		assert((int)pthread_getspecific(k1) == 24);
		assert((int)pthread_getspecific(k2) == 50);

		return (void*)123;
	}

	pthread_t tid;
	assert(pthread_create(&tid, NULL, &thread, NULL) == 0);
	assert(pthread_join(tid, NULL) == 0);
}

void test_pthreads_misc_key_destructors()
{
	static int dest1_called = 0;
	static int dest2_called = 0;

	void dest1(void *val)
	{
		dest1_called = 1;
		assert((int)val == 1);
	}
	void dest2(void *val)
	{
		dest2_called = 1;
		assert((int)val == 2);
	}

	void *func(void *arg)
	{
		pthread_key_t key1, key2;
		pthread_key_create(&key1, &dest1);
		pthread_key_create(&key2, &dest2);

		pthread_setspecific(key1, (void*)1);
		pthread_setspecific(key2, (void*)2);
		return NULL;
	}

	test_init();
	print_function_name();

	pthread_t tid;
	assert(pthread_create(&tid, NULL, &func, NULL) == 0);
	assert(pthread_join(tid, NULL) == 0);

	assert(dest1_called == 1);
	assert(dest2_called == 1);
}
