#include "shmac/shmac.h"
#include "shmac/hw_timer.h"
#include "shmac/irq.h"
#include "shmac/swi.h"

#include "debug.h"
#include "context.h"
#include "timer.h"

static int timer_number[MAX_CPUS];

static void enable_interrupts()
{
	*hw_timer_ctrl(timer_number[*TILEREG_CPUID]) &= ~TIMER_CTRL_ENABLE;
	*hw_timer_clr(timer_number[*TILEREG_CPUID]) = 1;

	asm("mrs r0, cpsr\n"
		"bic r0, r0, #0x80\n"
		"msr cpsr, r0\n"
		:
		:
		:"r0");
}
static void disable_interrupts()
{
	asm("mrs r0, cpsr\n"
	    "orr r0, r0, #0x80\n"
	    "msr cpsr, r0\n"
		:
		:
		:"r0");

	*hw_timer_ctrl(timer_number[*TILEREG_CPUID]) |= TIMER_CTRL_ENABLE;
}

static void timer_tick_handle(irq_regs_t *registers)
{
	irq_regs_t local_registers = *registers;
	enable_interrupts();
	schedule(&local_registers);
	disable_interrupts();
	*registers = local_registers;
}

static void software_interrupt_handle(char number, irq_regs_t *registers)
{
	// internally, schedule() expects PC to point to the instruction
	// _after_ the one we're returning to. hence the +/- 4
	registers->r[REG_PC] += 4;

	switch (number)
	{
		case SWI_PTHREAD_TRY_SCHEDULE:
			schedule(registers);
			break;
		case SWI_PTHREAD_MAIN_THREAD:
			schedule_main(registers);
			break;
		default:
			warn("Bug! unexpected swi: %d", number);
			break;
	}

	registers->r[REG_PC] -= 4;
}


/*
 * Timer setup
 */

void setup_timer()
{
	timer_number[*TILEREG_CPUID] = hw_timer_acquire();
	if (timer_number[*TILEREG_CPUID] == HW_TIMER_UNAVAILABLE) {
		warn("Failed to acquire timer!");
		return;
	} else {
		debug("acquired timer %d", timer_number[*TILEREG_CPUID]);
	}

    irq_timer_callback_set(timer_number[*TILEREG_CPUID], timer_tick_handle);

    // set interval
    *hw_timer_load(timer_number[*TILEREG_CPUID]) = TIMER_TICKS;

    // make it periodic
    *hw_timer_ctrl(timer_number[*TILEREG_CPUID]) = TIMER_CTRL_PERIODIC | TIMER_CTRL_SCALE_1;

    // we want to be able to forcibly activate the scheduler by interrupting manually. Mostly
    // just for performance reasons.
    swi_handler_register(SWI_PTHREAD_TRY_SCHEDULE, &software_interrupt_handle);

    // this is used to turn core 0's execution context into a task whenever it first
    // calls pthread_create. this to make core 0 schedulable along with the rest.
    swi_handler_register(SWI_PTHREAD_MAIN_THREAD, &software_interrupt_handle);
}

void stop_timer()
{
	irq_timer_callback_clear(timer_number[*TILEREG_CPUID]);
}

void start_timer()
{
	// reset value
	*hw_timer_load(timer_number[*TILEREG_CPUID]) = TIMER_TICKS;

	// re-enable timer
	*hw_timer_ctrl(timer_number[*TILEREG_CPUID]) |= TIMER_CTRL_ENABLE;

	// enable timer interrupts
    *IC0_IRQ_ENABLESET |= hw_timer_int_mask(timer_number[*TILEREG_CPUID]);
}

void pause_timer()
{
	*hw_timer_ctrl(timer_number[*TILEREG_CPUID]) &= ~TIMER_CTRL_ENABLE;
}

void continue_timer()
{
	*hw_timer_ctrl(timer_number[*TILEREG_CPUID]) |= TIMER_CTRL_ENABLE;
}
