#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <errno.h>
#include <pthread.h>

#include "shmac/shmac.h"
#include "shmac/swi.h"
#include "shmac/sleep.h"

#include "sched/sched.h"
#include "timer.h"
#include "core.h"
#include "core_tq.h"
#include "global_tq.h"
#include "event_log.h"
#include "task.h"
#include "worker.h"

static void pop_cleanup_stack(task_t *task)
{
    if (task != NULL) {
        int n = linked_list_size(&(task->cleanup_handlers));
        while (n--) {
            pthread_cleanup_pop(1);
        }
    }
}

void execute_task(task_t *task)
{
    // this function is called from timer interrupts when a new
    // task should be started. It does not have a complete stack
    // or valid lr, so it must block until reaped.

    func_t func = task->task;
    void *arg = task->arg;

    log_taskf(TASK_EXECUTING, task, 2, func, arg);
    task->ret = func(arg);

    log_task(TASK_EXITED, task);
    finish_task(task);
}

void finish_task(task_t *task)
{
    // pthread_key.c
    extern void pthread_keys_destroy();

    pthread_keys_destroy();
    pop_cleanup_stack(task);

    task->state = FINISHED;
    schedule_try();
    while (1);
}

void free_task(task_t *task)
{
    global_tq_remove(*(task->thread));

    if (task->stack_free) {
        free(task->attrs.stackaddr);
    }
    if (task->task_free) {
        free(task);
    }
}

static void *idle_func(void *_)
{
    while (1);
    return NULL;
}

static void schedule_idle_task(task_t *idle, sched_entity_t *se, pthread_t *thread, pthread_attr_t *att)
{
    att->stacksize = 0;
    att->stackaddr = NULL;
    pthread_attr_init(att);

    const struct sched_param sp = { .sched_priority = 0 };
    pthread_attr_setschedparam(att, &sp);
    pthread_attr_setschedpolicy(att, SCHED_OTHER);

    *thread              = *TILEREG_CPUID;

    idle->state          = QUEUED;               // task is still not running
    idle->cpu_no         = *TILEREG_CPUID;       // .. this one, in fact

    idle->thread         = thread;
    idle->attrs          = *att;                 // default attributes

    idle->stack_free     = 1;                    // stack is not heap-alloced
    idle->task_free      = 0;

    idle->task           = &idle_func;
    idle->arg            = NULL;
    idle->ret            = NULL;

    idle->cancel_type    = PTHREAD_CANCEL_ASYNCHRONOUS;
    idle->cancel_state   = PTHREAD_CANCEL_ENABLE;

    idle->waits_for          = NULL;
    idle->join_ll_item.value = NULL;
    idle->join_ll_item.prev  = NULL;
    idle->join_ll_item.next  = NULL;

    idle->pinned         = 1;
    idle->blockable      = 1;

    idle->cond_waiting      = 0;
    idle->cond_wait_timeout = 0;
    idle->barrier_waiting   = 0;
    idle->join_waiting      = 0;

    idle->epoch_count    = 0;

    for (int reg = 0; reg < ARM_REG_COUNT; reg++) {
        idle->register_set.r[reg] = 0;
    }
    idle->register_set.r[REG_SPSR] = 0x10;
    idle->register_set.r[REG_R0] = (int)&idle_func;

    sched_entity_init(se, idle);

    log_task(TASK_CREATED, idle);

    global_tq_insert(*(idle->thread), idle);
    core_tq_push(*TILEREG_CPUID, idle, LOCK);

    log_taskf(TASK_ASSIGNED, idle, 2, *TILEREG_CPUID, 0);
}

void pthreads_work_loop()
{
    if (*TILEREG_CPUID == 0)
        *IC0_IRQ_ENABLESET = INT_MASK_HOST;

    // create a timer object that will give us
    // regular interrupts to context switch
    setup_timer();

    // create a low-priority task that will never end.
    task_t *idle;
    sched_entity_t *se;
    pthread_t *idle_tid;
    pthread_attr_t *attrs;

    if (*TILEREG_CPUID == 0) {
        idle = malloc(sizeof(task_t) + sizeof(pthread_t) +
                sizeof(pthread_attr_t) + sizeof(sched_entity_t));
        se = (void*)idle + sizeof(task_t);
        idle_tid = (void*)se + sizeof(sched_entity_t);
        attrs = (void*)idle_tid + sizeof(pthread_t);

    } else {
        task_t _idle;
        sched_entity_t _se;
        pthread_t _idle_tid;
        pthread_attr_t _attrs;

        idle = &_idle;
        se = &_se;
        idle_tid = &_idle_tid;
        attrs = &_attrs;
    }

    schedule_idle_task(idle, se, idle_tid, attrs);

    if (*TILEREG_CPUID != 0) {
        // start executing the idle job (or something higher priority)
        start_timer();
    }

    // now we can start receiving tasks
    set_my_state(IDLE);

    // will just wait here until some task is started. we will
    // never return here, anyway.
    if (*TILEREG_CPUID != 0) {
        while (1);
    }
}
